/******************************************************************************
 *
 * Filename: engine.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  The execution engine.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "config.h"
#include "classfile.h"
#include "disasm.h"
#include "instructions.h"
#include "linker.h"
#include "util.h"
#include "memory.h"

#include "engine.h"


/* instruction peeks */
#define PEEKU1(p) (u1) (*(( p )+1))
#define PEEKU2(p) (u2) ((*(( p )+1) << 8) | *(( p )+2))
#define PEEKU4(p) (u4) ((*(( p )+1) << 24) | (*(( p )+2) << 16) | (*(( p )+3) << 8) | *(( p )+4))

#define INSTRUCTION if( instruction_debug ) printf

/* debug flags */
extern int verbose;
extern int instruction_debug;
extern int method_debug;
extern int field_debug;
extern int arithmetic_debug;
extern int instruction_count;

/* type check failure flag */
int halt = 0;

void (*callstack)( Frame*, int ) = NULL;
void (*instruction)( Frame* ) = NULL;

ObjRefList* objectPool = NULL;
ObjRefList* arrayPool = NULL;


int isFNaN( u4 v )
{
	return (((v >= 0xff800001) && (v <= 0xffffffff))
		|| ((v >= 0x7f800001) && (v <= 0x7fffffff)));
}

int isFpInf( u4 v )  { return ( v == 0x7f800000 ); }

int isFnInf( u4 v )  { return ( v == 0xff800000 ); }

int isDNaN( u4 vw[2] )
{
	int64* v = (int64*) vw; 
	return ( ((*v >= 0xff80000000000001L) && (*v <= 0xffffffffffffffffL) && (*v != 0xfff0000000000000L))
		|| ((*v >= 0x7ff8000000000000L) && (*v <= 0x7fffffffffffffffL) && (*v != 0x7ff0000000000000L)) );
}

int isDpInf( u4 vw[2] )
{
	int64* v = (int64*) vw; 
	return ( *v == 0x7ff0000000000000L );
}

int isDnInf( u4 vw[2] )
{
	int64* v = (int64*) vw;
	return ( *v == 0xfff0000000000000L );
}

u4 fNaN = 0xff800001;
u4 fpInf = 0x7f800000;
u4 fnInf = 0xff800000;

u4 dNaN[2] = { 0x00000000, 0x7ff80000  };
u4 dpInf[2] = { 0x00000000, 0x7ff00000  };
u4 dnInf[2] = { 0x00000000, 0xfff00000  };

#define MAX_INTEGER 0x7fffffff
#define MIN_INTEGER 0x80000000
#define MAX_LONG 0x7fffffffffffffffL
#define MIN_LONG 0x8000000000000000L


/*****************************************************************************/

/* check to see if one class is a subclass of another */
int subclassOf( const ClassFile* cf, const ClassFile* compare )
{
	u2 n;

	if( !cf || !compare )
		return 0;

	/*if( verbose )
	{
		char exception[255];
		char catcher[255];
		sprintClassName( exception, (ClassFile*) cf, cf->this_class );
		sprintClassName( catcher, (ClassFile*) compare, compare->this_class );
		printf( "Checking \"%s\" subclassOf \"%s\"\n", exception, catcher );
	}*/

	if( ((u4) cf) == ((u4) compare) )
		return 1;

	if( cf->super_class && CHECKCONSTANT( cf, cf->super_class, CONSTANT_Class )
		&& subclassOf( ((CONSTANT_Class_info*) cf->constant_pool[cf->super_class])->cf, compare ) )
		return 1;

	for( n = 0; n < cf->interfaces_count; n++ )
	{
		if( cf->interfaces[n] && CHECKCONSTANT( cf, cf->interfaces[n]->class_index, CONSTANT_Class )
			&& subclassOf( ((CONSTANT_Class_info*) cf->constant_pool[cf->interfaces[n]->class_index])->cf, compare ) )
			return 1;
	}

	return 0;
}


#define INITIALISE( cf, cfp ) Stack stack; stack.ex = NULL; stack.fp = NULL; \
	if( initialise( cf, &stack ) == 2 ) { \
		if( stack.fp && stack.fp->ex ) (cfp)->ex = stack.fp->ex; \
		else (cfp)->ex = stack.ex; return 2; }


int getFieldType( field_info* fp )
{
	if( CHECKCONSTANT( fp->cf, fp->descriptor_index, CONSTANT_Utf8 ) )
	{
		CONSTANT_Utf8_info *utf8 = (CONSTANT_Utf8_info*) fp->cf->constant_pool[fp->descriptor_index];
	
		switch( utf8->bytes[0] )
		{
		case 'B': return T_INT; /* All byte fields are handled as int's */
		case 'C': return T_INT; /* All char fields are handled as int's */
		case 'D': return T_DOUBLE;
		case 'F': return T_FLOAT;
		case 'I': return T_INT;
		case 'J': return T_LONG;
		case 'L': return T_OBJREF;
		case 'S': return T_INT; /* All short fields are handled as int's */
		case 'Z': return T_INT; /* All boolean fields are handled as int's */
		case '[': return T_OBJREF;
		}
	}

	return T_ANY;
}


u4 popCheck( Frame* fp, u1 type )
{
	(fp->tsp)--; (fp->sp)--;
	if( *(fp->tsp) != type )
	{
		DEBUG( printf( "ERROR: popCheck, type %d != %d\n", *(fp->tsp), type ); )
		halt = 1;
	}
	return *(fp->sp);
}


u4 popCheck1( Frame* fp, u1 type )
{
	(fp->tsp)--; (fp->sp)--;
	if( *(fp->tsp) != type )
	{
		DEBUG( printf( "ERROR: popCheck1, type %d != %d\n", *(fp->tsp), type ); )
		halt = 1;
	}
	return *(fp->sp);
}


u4 popCheckAny( Frame* fp )
{
	(fp->tsp)--; (fp->sp)--;
	if( *(fp->tsp) > T_LAST )
	{
		DEBUG( printf( "ERROR: popCheckAny, type %d > T_LAST\n", *(fp->tsp) ); )
		halt = 1;
	}
#ifdef _DEBUG
	else if( *(fp->tsp) == 0 )
	{
		printf( "WARNING: popCheckAny, type %d == 0\n", *(fp->tsp) );
	}
#endif
	return *((fp)->sp);
}


void pushCheckLocal( Frame* fp, u2 index, u1 type )
{
	if( fp->tlocal[index] != type )
	{
		DEBUG( printf( "ERROR: pushCheckLocal, type %d != %d\n", fp->tlocal[index], type ); )
		halt = 1;
	}
	*(fp->sp) = fp->local[index]; (fp->sp)++;
	*(fp->tsp) = type; (fp->tsp)++;
}


void pushCheckLocal2( Frame* fp, u2 index, u1 type0, u1 type1 )
{
	if( fp->tlocal[index+1] != type0 )
	{
		DEBUG( printf( "ERROR: pushCheckLocal2 (1), type %d != %d\n", fp->tlocal[index+1], type0 ); )
		halt = 1;
	}
	if( fp->tlocal[index] != type1 )
	{
		DEBUG( printf( "ERROR: pushCheckLocal2 (0), type %d != %d\n", fp->tlocal[index], type0 ); )
		halt = 1;
	}
	*(fp->sp) = fp->local[index]; (fp->sp)++;
	*(fp->tsp) = ( fp->tlocal[index] ? fp->tlocal[index] : type1 ); (fp->tsp)++;
	*(fp->sp) = fp->local[index+1]; (fp->sp)++;
	*(fp->tsp) = ( fp->tlocal[index+1] ? fp->tlocal[index+1] : type0 ); (fp->tsp)++;
}


void popCheckLocal( Frame* fp, u2 index, u1 type )
{
	(fp->tsp)--; (fp->sp)--;
	if( *(fp->tsp) != type )
	{
		DEBUG( printf( "ERROR: popCheckLocal, stack corrupt (type %d != %d)\n", fp->tlocal[index], type ); )
		halt = 1;
	}
	fp->local[index] = *(fp->sp);
	fp->tlocal[index] = type;
}


void popCheckLocal2( Frame* fp, u2 index, u1 type0, u1 type1 )
{
	(fp->tsp)--; (fp->sp)--;
	if( *(fp->tsp) != type0 )
	{
		DEBUG( printf( "ERROR: popCheckLocal2 (0), stack corrupt (type %d != %d)\n", fp->tlocal[index+1], type0 ); )
		halt = 1;
	}
	fp->local[index+1] = *(fp->sp);
	fp->tlocal[index+1] = type0;

	(fp->tsp)--; (fp->sp)--;
	if( *(fp->tsp) != type1 )
	{
		DEBUG( printf( "ERROR: popCheckLocal2 (1), stack corrupt (type %d != %d)\n", fp->tlocal[index], type1 ); )
		halt = 1;
	}
	fp->local[index] = *(fp->sp);
	fp->tlocal[index] = type1;
}


ObjRef* getLocalRetAddr( Frame* fp, u2 index )
{
	if( fp->tlocal[index] != T_RETADDR )
	{
		DEBUG( printf( "ERROR: getLocalObjRef, type %d != %d\n", fp->tlocal[index], T_OBJREF ); )
		halt = 1;
	}
	return (ObjRef*) fp->local[index];
}


int32 getLocalInt( Frame* fp, u2 index )
{
	if( fp->tlocal[index] && (fp->tlocal[index] != T_INT) )
	{
		DEBUG( printf( "ERROR: getLocalInt, type %d != %d\n", fp->tlocal[index], T_INT ); )
		halt = 1;
	}
	return (int32) fp->local[index];
}


/******************************* Instructions ********************************/

/******************************************************************************
 * A
 *****************************************************************************/

#include "object.c"


int anewarray( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;
	ObjRef* objref;
	int32 length;

	INSTRUCTION( "%p: anewarray %d\n", fp->pc, index );

	POP_INT( fp, length );

	if( length < 0 )
	{
		char buf[256];
		sprintf( buf, "%d", (int) length );
		return newCompleteException( stp, NegativeArraySizeException, buf );
	}
	
	objref = newObjectRefAndArray( T_OBJREF, length, sizeof(ObjRef*) );
	PUSH_OBJREF( fp, objref );

	/* add to array pool */
	addToList( objref, &arrayPool );
	return 0;
}


int arraylength( Stack* stp )
{
	Frame* fp = stp->fp;
	ObjRef* objref;

	INSTRUCTION( "%p: arraylength\n", fp->pc );

	POP_OBJREF( fp, objref );

	if( !objref )
		return newCompleteException( stp, NullPointerException, "arraylength" );

	if( (objref->type == 0) && !objref->cf )
		return newCompleteException( stp, InternalError, "Corrupt object reference!" );

	PUSH_INT( fp, objref->length );
	return 0;
}


int athrow( Stack* stp )
{
	Frame* fp = stp->fp;
	ObjRef* objref;

	INSTRUCTION( "%p: athrow\n", fp->pc );

	POP_OBJREF( fp, objref );

	if( !objref )
		return newCompleteException( stp, NullPointerException, "athrow" );

	fp->ex = (Exception*) javavm_malloc( sizeof( Exception ) );
	fp->ex->kind = JavaException;
	fp->ex->message = NULL;
	fp->ex->ex = objref;
	return 0;
}


/******************************************************************************
 * B
 *****************************************************************************/

#include "boolean.c"


/******************************************************************************
 * C
 *****************************************************************************/

#include "char.c"


int checkcast( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;
	ObjRef* objref;

	INSTRUCTION( "%p: checkcast %d\n", fp->pc, index );

	POP_OBJREF( fp, objref );

	if( !objref )
	{
		PUSH_OBJREF( fp, NULL );
		return 0;
	}

	if( (!objref->cf && !objref->type) || (objref->cf && objref->type) )
		return newCompleteException( stp, InternalError, "Corrupt object reference" );

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Class ) )
	{
		CONSTANT_Class_info* ci = (CONSTANT_Class_info*) fp->cf->constant_pool[index];

		/* array cast check */
		if( CHECKCONSTANT( fp->cf, ci->name_index, CONSTANT_Utf8 ) )
		{
			CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) fp->cf->constant_pool[ci->name_index];

			if( utfp->length && utfp->bytes && (utfp->bytes[0] == '[') )
			{
				if( !objref->cf && objref->type )
				{
					PUSH_OBJREF( fp, objref );
					return 0;
				}
				else
				{
					char buf[255], dcbuf[255], acbuf[255];
					sprintClassName( dcbuf, objref->cf, objref->cf->this_class );
					sprintClassName( acbuf, ci->cf, ci->cf->this_class );
					sprintf( buf, "casting from %s to %s", dcbuf, acbuf );
					return newCompleteException( stp, ClassCastException, buf );
				}
			}
		}

		if( !ci->cf )
			linkClass( fp->cf, index, stp );

		if( !ci->cf )
		{
			if( !objref->cf )
			{
				PUSH_OBJREF( fp, objref );
				stp->ex = NULL;
				if( stp->fp ) stp->fp->ex = NULL;
				return 0;
			}

			return stp->retcode; /* linkClass looks after exception generation */
		}

		if( subclassOf( objref->cf, ci->cf ) )
		{
			PUSH_OBJREF( fp, objref );
			return 0;
		}
		else
		{
			char buf[255], dcbuf[255], acbuf[255];
			sprintClassName( dcbuf, objref->cf, objref->cf->this_class );
			sprintClassName( acbuf, ci->cf, ci->cf->this_class );
			sprintf( buf, "casting from %s to %s", dcbuf, acbuf );
			return newCompleteException( stp, ClassCastException, buf );
		}
	}

	return newCompleteException( stp, InternalError, "Unable to extract classinfo during cast" );
}


/******************************************************************************
 * D
 *****************************************************************************/

#include "double.c"

#include "stackops.c"


/******************************************************************************
 * F
 *****************************************************************************/

#include "float.c"


/******************************************************************************
 * G
 *****************************************************************************/

void goto_inst( Frame* fp, u2 incr )
{
	INSTRUCTION( "%p: goto %d\n", fp->pc, (int16) incr );
	fp->pc += (int16) incr;
}


void goto_w( Frame* fp, u4 incr )
{
	INSTRUCTION( "%p: goto_w 0x%08x\n", fp->pc, (int32) incr );
	fp->pc += (int32) incr;
}


/******************************************************************************
 * I
 *****************************************************************************/

#include "integer.c"

#include "if.c"


int instanceof( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;
	ObjRef* objref;

	INSTRUCTION( "%p: instanceof %d\n", fp->pc, index );

	POP_OBJREF( fp, objref );

	if( !objref || !objref->cf )
	{
		PUSH_INT( fp, 0 );
		return 0;
	}

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Class ) )
	{
		CONSTANT_Class_info* ci = (CONSTANT_Class_info*) fp->cf->constant_pool[index];

		if( !ci->cf )
			linkClass( fp->cf, index, stp );

		if( !ci->cf )
			return stp->retcode; /* linkClass looks after exception generation */

		PUSH_INT( fp, subclassOf( objref->cf, ci->cf ) );
		return 0;
	}

	PUSH_INT( fp, 0 );
	return 0;
}


/******************************************************************************
 * J
 *****************************************************************************/

void jsr( Frame* fp, u2 incr )
{
	INSTRUCTION( "%p: jsr %d\n", fp->pc, incr );

	PUSH_RETADDR( fp, fp->pc + 3 );
	fp->pc += (int16) incr;
}


void jsr_w( Frame* fp, u4 incr )
{
	INSTRUCTION( "%p: jsr_w 0x%08x\n", fp->pc, incr );

	PUSH_RETADDR( fp, fp->pc + 3 );
	fp->pc += (int32) incr;
}


/******************************************************************************
 * L
 *****************************************************************************/

#include "long.c"


int ldc( Stack* stp, u2 index )
{
	Frame *fp = stp->fp;

	INSTRUCTION( "%p: ldc %d\n", fp->pc, index );

	if( (index > 0) && (index < fp->cf->constant_pool_count) )
	{
		switch( fp->cf->constant_pool[index]->tag )
		{
		case CONSTANT_String:
			{
			CONSTANT_String_info* sp = (CONSTANT_String_info*) fp->cf->constant_pool[index];
			if( CHECKCONSTANT( fp->cf, sp->string_index, CONSTANT_Utf8 ) )
			{
				if( !sp->objref )
				{
					CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) fp->cf->constant_pool[sp->string_index];
					sp->objref = newJavaLangString( utfp->bytes, utfp->length );

					if( !sp->objref )
						return newCompleteException( stp, InternalError, "ldc failed - failed to create string" );
				}

				PUSH_OBJREF( fp, sp->objref );
				return 0;
			}
			}
			break;
		case CONSTANT_Integer:
			{
			CONSTANT_Integer_info* ip = (CONSTANT_Integer_info*) fp->cf->constant_pool[index];
			PUSH_INT( fp, ip->bytes );
			return 0;
			}
			break;
		case CONSTANT_Float:
			{
			CONSTANT_Float_info* ip = (CONSTANT_Float_info*) fp->cf->constant_pool[index];
			PUSH_FLOAT( fp, ip->bytes );
			return 0;
			}
			break;
		default:
			{
			char buf[255];
			sprintf( buf, "Unsupported ldc type (%d)", fp->cf->constant_pool[index]->tag );
			return newCompleteException( stp, InternalError, buf );
			}
		}
	}

	return newCompleteException( stp, InternalError, "ldc failed - bad constant pool" );
}


int ldc2_w( Stack* stp, u2 index )
{
	Frame *fp = stp->fp;

	INSTRUCTION( "%p: ldc2_w %d\n", fp->pc, index );

	if( (index > 0) && (index < fp->cf->constant_pool_count) )
	{
		switch( fp->cf->constant_pool[index]->tag )
		{
		case CONSTANT_Long:
			{
			u4 value[2];
			CONSTANT_Long_info* ip = (CONSTANT_Long_info*) fp->cf->constant_pool[index];
			value[1] = ip->high_bytes;
			value[0] = ip->low_bytes;
			PUSH_LONG( fp, value );
			return 0;
			}
			break;
		case CONSTANT_Double:
			{
			u4 value[2];
			CONSTANT_Double_info* ip = (CONSTANT_Double_info*) fp->cf->constant_pool[index];
			value[1] = ip->high_bytes;
			value[0] = ip->low_bytes;
			PUSH_DOUBLE( fp, value );
			return 0;
			}
			break;
		default:
			{
			char buf[255];
			sprintf( buf, "Unsupported ldc2_w type (%d)", fp->cf->constant_pool[index]->tag );
			return newCompleteException( stp, InternalError, buf );
			}
		}
	}

	return newCompleteException( stp, InternalError, "ldc2_w failed - bad constant pool" );
}


void lookupswitch( Frame* fp )
{
	int32 key, match, offset = 0, def;
	u4 pcoff = fp->pc - fp->min_pc;
	u4 n, count;
	u1* pc;

	pcoff = ((pcoff)/4)*4 + 3;
	pc = fp->min_pc + pcoff;

	def = (int32) PEEKU4( pc ); pc += 4;
	count = PEEKU4( pc ); pc += 4;

	INSTRUCTION( "%p: lookupswitch %d %d\n", fp->pc, (int) def, (int) count );

	POP_INT( fp, key );

	for( n = 0; n < count; n++ )
	{
		match  = (int32) PEEKU4( pc ); pc += 4;
		offset = (int32) PEEKU4( pc ); pc += 4;

		if( key == match )
			break;
	}

	if( n >= count )
		offset = def;

	fp->pc += offset;
}


int getfield( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;
	ObjRef* objref;

	INSTRUCTION( "%p: getfield %d\n", fp->pc, index );

	POP_OBJREF( fp, objref );

	if( !objref )
		return newCompleteException( stp, NullPointerException, "getfield" );

	if( objref->cf && !objref->cf->initialised )
	{
		INITIALISE( objref->cf, fp );
	}

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Fieldref ) )
	{
		CONSTANT_Fieldref_info* frp = (CONSTANT_Fieldref_info*) fp->cf->constant_pool[index];

		if( !frp->info )
		{
			linkFieldRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkFieldRef looks after exception generation */
		}

		if( frp->info && frp->info->cf && objref->data )
		{
#ifdef _DEBUG
			if( field_debug )
			{
				char buf[255];
				sprintFieldRef( buf, fp->cf, index );
				printf( "Get Field %s (0x%p, [%d], 0x%08x)\n",
					buf, objref->data, (int) frp->info->offset, objref->data[frp->info->offset] );
			}
#endif /* _DEBUG */

			switch( getFieldType( frp->info ) )
			{
			case T_OBJREF:
				PUSH_OBJREF( fp, objref->data[frp->info->offset] );
				break;
			case T_INT:
				PUSH_INT( fp, objref->data[frp->info->offset] );
				break;
			case T_FLOAT:
				PUSH_FLOAT( fp, objref->data[frp->info->offset] );
				break;
			case T_LONG:
				{
				u4 value[2];
				value[0] = objref->data[frp->info->offset];
				value[1] = objref->data[frp->info->offset+1];
				PUSH_LONG( fp, value );
				}
				break;
			case T_DOUBLE:
				{
				u4 value[2];
				value[0] = objref->data[frp->info->offset];
				value[1] = objref->data[frp->info->offset+1];
				PUSH_DOUBLE( fp, value );
				}
				break;
			default:
				printf( "ERROR: untyped field being pushed\n" );
				return newCompleteException( stp, InternalError, "untyped field being pushed" );
			}
			return 0;
		}
		else
			return newNoSuchMethodError( stp, fp->cf, index );
	}

	return newNoSuchMethodError( stp, fp->cf, index );
}


int getstatic( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;

	INSTRUCTION( "%p: getstatic %d\n", fp->pc, index );

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Fieldref ) )
	{
		CONSTANT_Fieldref_info* frp = (CONSTANT_Fieldref_info*) fp->cf->constant_pool[index];

		if( !frp->info )
		{
			linkFieldRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkFieldRef looks after exception generation */
		}

		if( frp->info && frp->info->cf && !frp->info->cf->initialised )
		{
			INITIALISE( frp->info->cf, fp );
		}

		if( frp->info && frp->info->cf && frp->info->cf->static_data )
		{
#ifdef _DEBUG
			if( field_debug )
			{
				char buf[255];
				sprintFieldRef( buf, fp->cf, index );
				printf( "Get Static Field %s (0x%p, [%d], 0x%08x)\n",
					buf, frp->info->cf->static_data, (int) frp->info->offset,
					frp->info->cf->static_data[frp->info->offset] );
			}
#endif /* _DEBUG */

			switch( getFieldType( frp->info ) )
			{
			case T_OBJREF:
				PUSH_OBJREF( fp, frp->info->cf->static_data[frp->info->offset] );
				break;
			case T_INT:
				PUSH_INT( fp, frp->info->cf->static_data[frp->info->offset] );
				break;
			case T_FLOAT:
				PUSH_FLOAT( fp, frp->info->cf->static_data[frp->info->offset] );
				break;
			case T_LONG:
				{
				u4 value[2];
				value[0] = frp->info->cf->static_data[frp->info->offset];
				value[1] = frp->info->cf->static_data[frp->info->offset+1];
				PUSH_LONG( fp, value );
				}
				break;
			case T_DOUBLE:
				{
				u4 value[2];
				value[0] = frp->info->cf->static_data[frp->info->offset];
				value[1] = frp->info->cf->static_data[frp->info->offset+1];
				PUSH_DOUBLE( fp, value );
				}
				break;
			default:
				printf( "ERROR: untyped static field being pushed\n" );
				return newCompleteException( stp, InternalError, "untyped field being pushed" );
			}
			return 0;
		}
	}

	return newNoSuchMethodError( stp, fp->cf, index );
}


int putfield( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;
	ObjRef* objref;
	u4 value, value1 = 0;
	u1 stype, stype1, ftype;

	INSTRUCTION( "%p: putfield %d\n", fp->pc, index );

	POP_ANY_TYPE( fp, value, stype );

	switch( stype )
	{
	case T_LONG:
	case T_DOUBLE:
		POP_ANY_TYPE( fp, value1, stype1 );
		break;
	case T_LONG1:
	case T_DOUBLE1:
		halt = 1;
		break;
	}

	POP_OBJREF( fp, objref );

	if( !objref )
		return newCompleteException( stp, NullPointerException, "putfield" );

	if( objref->cf && !objref->cf->initialised )
	{
		INITIALISE( objref->cf, fp );
	}

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Fieldref ) )
	{
		CONSTANT_Fieldref_info* frp = (CONSTANT_Fieldref_info*) fp->cf->constant_pool[index];

		if( !frp->info )
		{
			linkFieldRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkFieldRef looks after exception generation */
		}

		if( frp->info && frp->info->cf && objref->data )
		{
#ifdef _DEBUG
			if( field_debug )
			{
				char buf[255];
				sprintFieldRef( buf, fp->cf, index );
				printf( "Put Field %s (0x%p, [%d], 0x%08x)\n",
					buf, objref->data, (int) frp->info->offset, value );
			}
#endif /* _DEBUG */

			ftype = getFieldType( frp->info );
			if( ftype && stype && (ftype != stype) )
				halt = 1;
			objref->data[frp->info->offset] = value;

			switch( stype )
			{
			case T_LONG:
			case T_DOUBLE:
				objref->data[frp->info->offset+1] = value1;
				break;
			}

			return 0;
		}
	}

	return newNoSuchMethodError( stp, fp->cf, index );
}


int putstatic( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;
	u4 value, value1 = 0;
	u1 stype, stype1, ftype;

	INSTRUCTION( "%p: putstatic %d\n", fp->pc, index );

	POP_ANY_TYPE( fp, value, stype );

	switch( stype )
	{
	case T_LONG:
	case T_DOUBLE:
		POP_ANY_TYPE( fp, value1, stype1 );
		break;
	case T_LONG1:
	case T_DOUBLE1:
		halt = 1;
		break;
	}

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Fieldref ) )
	{
		CONSTANT_Fieldref_info* frp = (CONSTANT_Fieldref_info*) fp->cf->constant_pool[index];

		if( !frp->info )
		{
			linkFieldRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkFieldRef looks after exception generation */
		}

		if( frp->info && frp->info->cf && !frp->info->cf->initialised )
		{
			INITIALISE( frp->info->cf, fp );
		}

		if( frp->info && frp->info->cf && frp->info->cf->static_data )
		{
#ifdef _DEBUG
			if( field_debug )
			{
				char buf[255];
				sprintFieldRef( buf, fp->cf, index );
				printf( "Put Static Field %s (0x%p, [%d], 0x%08x)\n",
					buf, frp->info->cf->static_data, (int) frp->info->offset, value );
			}
#endif /* _DEBUG */

			ftype = getFieldType( frp->info );
			if( ftype && stype && (ftype != stype) )
				halt = 1;

			frp->info->cf->static_data[frp->info->offset] = value;

			switch( stype )
			{
			case T_LONG:
			case T_DOUBLE:
				frp->info->cf->static_data[frp->info->offset+1] = value1;
				break;
			}

			return 0;
		}
	}

	return newNoSuchMethodError( stp, fp->cf, index );
}


int invokeinterface( Stack* stp, u2 index, u1 args )
{
	Frame* fp = stp->fp;

	INSTRUCTION( "%p: invokeinterface %d %d\n", fp->pc, index, args );

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_InterfaceMethodref ) )
	{
		CONSTANT_InterfaceMethodref_info* mp = (CONSTANT_InterfaceMethodref_info*) fp->cf->constant_pool[index];
		method_info* mi;
		ObjRef* obj;
		u1 tobj;

		if( !mp->info )
		{
			linkInterfaceMethodRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkInterfaceMethodRef looks after exception generation */
		}

		if( !mp->info )
			newNoSuchMethodErrorIface( stp, fp->cf, index );

		mi = mp->info;

		if( (args - 1) != mi->args )
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, fp->cf, index );
			printf( "WARNING: invokeinterface argument counts don't match (%d != %d) for %s\n", mi->args, args, buf );
		}

		obj = (ObjRef*) *(fp->sp - mi->args - 1);
		tobj = *(fp->tsp - mi->args - 1);
		if( tobj != T_OBJREF )
		{
			halt = 1;
			return 2;
		}

		if( obj && (obj->cf != mi->cf) )
		{
			if( !obj->cf || !mi->cf )
				return newCompleteException( stp, InternalError, "Corrupt object reference" );

			/* should check that this->cf is a sub-class of mi->cf */
			if( obj->cf->virtuals && (mi != obj->cf->virtuals[mi->index]) )
			{
				int m;

				for( m = 0; (m < obj->cf->virtuals_count)
					&& !( compareUTF8( mi->cf, mi->name_index, obj->cf->virtuals[m]->cf, obj->cf->virtuals[m]->name_index )
						&& compareUTF8( mi->cf, mi->descriptor_index, obj->cf->virtuals[m]->cf, obj->cf->virtuals[m]->descriptor_index ) );
					m++ );

				if( m < obj->cf->virtuals_count )
				{
					/*char mbuf[255], vbuf[255];
					sprintMethodInfo( mbuf, mi );
					sprintMethodInfo( vbuf, obj->cf->virtuals[m] );
					printf( "Virtual method %s overridden by %s\n", mbuf, vbuf );*/

					mi = obj->cf->virtuals[m];
				}
				else
				{
					/*###*/char mbuf[255], vbuf[255];
					sprintMethodInfo( mbuf, mi );
					sprintMethodInfo( vbuf, obj->cf->virtuals[mi->index] );
					printf( "ERROR: Can't find override for %s using %s\n", mbuf, vbuf ); 
					mi = obj->cf->virtuals[mi->index];
				}
			}
		}

		if( mi->access_flags & ACC_NATIVE )
		{
			NativeWrapper method;
			
#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255];
				sprintInterfaceMethodRef( buf, fp->cf, index );
				printf( "Native method %s\n", buf );
			}
#endif /* _DEBUG */
			
			if( !mi->nativestr )
				return newUnsatisfiedLinkErrorIface( stp, fp->cf, index );
			
			method = mi->nativestr->wrapper;
			return (*method)( stp, mi->cf->classobj, mi->nativestr->arg );
		}
		else
		{
			/* new stack frame */
			Frame* frame = newStackFrame( mi, fp );
			ObjRef* objref;
			u1 n;
			
			if( !frame )
				return newUnsatisfiedLinkErrorIface( stp, fp->cf, index );
			
			/* pop args using instructions arg count*/
			for( n = args-1; n > 0; n-- )
			{
				POP_ANY_TYPE( fp, frame->local[n], frame->tlocal[n] );
			}
			
			/* pop objref into this (local[0]) */
			POP_OBJREF( fp, objref );

			if( halt )
				return 1;

			if( !objref )
			{
				freeStackFrame( frame, mi );
				return newCompleteException( stp, NullPointerException, "invokeinterface" );
			}

			if( !objref->cf || objref->type || (objref->length != 1) )
			{
				freeStackFrame( frame, mi );
				return newCompleteException( stp, InternalError, "Corrupt object reference" );
			}

			if( !objref->cf->initialised )
			{
				INITIALISE( objref->cf, fp );
			}

			PUT_LOBJREF( frame, 0, objref );

			/* Commit */
			fp->prev = frame;
			stp->fp = frame;

#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255];
				sprintInterfaceMethodRef( buf, fp->cf, index );
				printf( "Java method %s\n", buf );
			}
#endif /* _DEBUG */
			
			if( callstack )
				(*callstack)( frame, 1 /* push */ );
			
			return 0;
		}
	}

	return newCompleteException( stp, InternalError, "Bad InterfaceMethodRef" );
}


int invokespecial( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;

	INSTRUCTION( "%p: invokespecial %d\n", fp->pc, index );

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Methodref ) )
	{
		CONSTANT_Methodref_info* mp = (CONSTANT_Methodref_info*) fp->cf->constant_pool[index];

		if( !mp->info )
		{
			linkMethodRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkMethodRef looks after exception generation */
		}

		if( !mp->info )
			newNoSuchMethodError( stp, fp->cf, index );

		if( mp->info->access_flags & ACC_NATIVE )
		{
			NativeWrapper method;

#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255];
				sprintMethodRef( buf, fp->cf, index );
				printf( "Native method %s\n", buf );
			}
#endif /* _DEBUG */

			if( !mp->info->nativestr )
				return newUnsatisfiedLinkError( stp, fp->cf, index );

			method = mp->info->nativestr->wrapper;
			return (*method)( stp, mp->info->cf->classobj, mp->info->nativestr->arg );
		}
		else
		{
			/* new frame */
			Frame* frame = newStackFrame( mp->info, fp );
			ObjRef *objref;
			u1 n;

			if( !frame )
				return newUnsatisfiedLinkError( stp, fp->cf, index );

			/* pop args */
			for( n = mp->info->args; n > 0; n-- )
			{
				POP_ANY_TYPE( fp, frame->local[n], frame->tlocal[n] );
			}

			/* pop objref into this (local[0]) */
			POP_OBJREF( fp, objref );

			if( halt )
				return 1;

			if( !objref )
			{
				freeStackFrame( frame, mp->info );
				return newCompleteException( stp, NullPointerException, "invokespecial" );
			}

			if( !objref->cf || objref->type || (objref->length != 1) )
			{
				freeStackFrame( frame, mp->info );
				return newCompleteException( stp, InternalError, "Corrupt object reference" );
			}

			if( !objref->cf->initialised )
			{
				INITIALISE( objref->cf, fp );
			}

			PUT_LOBJREF( frame, 0, objref );

			/* Commit */
			fp->prev = frame;
			stp->fp = frame;

#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255];
				sprintMethodRef( buf, fp->cf, index );
				printf( "Java method %s\n", buf );
			}
#endif /* _DEBUG */

			if( callstack )
				(*callstack)( frame, 1 /* push */ );

			return 0;
		}
	}

	return newCompleteException( stp, InternalError, "Bad MethodRef" );
}


int invokestatic( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;

	INSTRUCTION( "%p: invokestatic %d\n", fp->pc, index );

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Methodref ) )
	{
		CONSTANT_Methodref_info* mp = (CONSTANT_Methodref_info*) fp->cf->constant_pool[index];

		if( !mp->info )
		{
			linkMethodRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkMethodRef looks after exception generation */
		}

		if( !mp->info )
			newNoSuchMethodError( stp, fp->cf, index );

		if( mp->info->access_flags & ACC_NATIVE )
		{
			NativeWrapper method;

#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255];
				sprintMethodRef( buf, fp->cf, index );
				printf( "Native method %s\n", buf );
			}
#endif /* _DEBUG */

			if( !mp->info->nativestr )
				return newUnsatisfiedLinkError( stp, fp->cf, index );

			method = mp->info->nativestr->wrapper;
			return (*method)( stp, mp->info->cf->classobj, mp->info->nativestr->arg );
		}
		else
		{
			/* new frame */
			Frame* frame = newStackFrame( mp->info, fp );
			u1 n;

			if( !frame )
				return newUnsatisfiedLinkError( stp, fp->cf, index );

			/* pop args */
			for( n = mp->info->args; n > 0; n-- )
			{
				POP_ANY_TYPE( fp, frame->local[n-1], frame->tlocal[n-1] );
			}

			if( !mp->info->cf->initialised )
			{
				INITIALISE( mp->info->cf, fp );
			}

			/* Commit */
			fp->prev = frame;
			stp->fp = frame;

#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255];
				sprintMethodRef( buf, fp->cf, index );
				printf( "Java method %s\n", buf );
			}
#endif /* _DEBUG */

			if( callstack )
				(*callstack)( frame, 1 /* push */ );

			return 0;
		}
	}

	return newCompleteException( stp, InternalError, "Bad MethodRef" );
}


int invokevirtual( Stack* stp, u2 index )
{
	Frame* fp = stp->fp;

	INSTRUCTION( "%p: invokevirtual %d\n", fp->pc, index );

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Methodref ) )
	{
		CONSTANT_Methodref_info* mp = (CONSTANT_Methodref_info*) fp->cf->constant_pool[index];
		method_info* mi;
		ObjRef* obj;
		u1 tobj;

		if( !mp->info )
		{
			linkMethodRef( fp->cf, index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) )
				return stp->retcode; /* linkMethodRef looks after exception generation */
		}

		if( !mp->info )
			newNoSuchMethodError( stp, fp->cf, index );

		mi = mp->info;

		obj = (ObjRef*) *(fp->sp - mi->args - 1);
		tobj = *(fp->tsp - mi->args - 1);
		if( tobj != T_OBJREF )
		{
			halt = 1;
			return 2;
		}

		if( obj && (obj->cf != mi->cf) )
		{
			if( !obj->cf || !mi->cf )
				return newCompleteException( stp, InternalError, "Corrupt object reference" );

			/* should check that this->cf is a sub-class of mi->cf */
			if( obj->cf->virtuals && (mi != obj->cf->virtuals[mi->index]) )
			{
				/*char mbuf[255], vbuf[255];
				sprintMethodInfo( mbuf, mi );
				sprintMethodInfo( vbuf, obj->cf->virtuals[mi->index] );
				printf( "Virtual method %s overridden by %s\n", mbuf, vbuf );*/

				mi = obj->cf->virtuals[mi->index];
			}
		}

		if( mi->access_flags & ACC_NATIVE )
		{
			NativeWrapper method;

#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255];
				sprintMethodInfo( buf, mi );
				printf( "Native method %s\n", buf );
			}
#endif /* _DEBUG */
			
			if( !mi->nativestr )
				return newUnsatisfiedLinkError( stp, fp->cf, index );
			
			method = mi->nativestr->wrapper;
			return (*method)( stp, mi->cf->classobj, mi->nativestr->arg );
		}
		else
		{
			/* new stack frame */
			Frame* frame = newStackFrame( mi, fp );
			ObjRef* objref;
			u1 n;
			
			if( !frame )
				return newUnsatisfiedLinkError( stp, fp->cf, index );
			
			/* pop args */
			for( n = mi->args; n > 0; n-- )
			{
				POP_ANY_TYPE( fp, frame->local[n], frame->tlocal[n] );
			}
			
			/* pop objref into this (local[0]) */
			POP_OBJREF( fp, objref );

			if( halt )
				return 1;

			if( !objref )
			{
				freeStackFrame( frame, mi );
				return newCompleteException( stp, NullPointerException, "invokevirtual" );
			}

			if( !objref->cf || objref->type || (objref->length != 1) )
			{
				freeStackFrame( frame, mi );
				return newCompleteException( stp, InternalError, "Corrupt object reference" );
			}

			if( !objref->cf->initialised )
			{
				INITIALISE( objref->cf, fp );
			}

			PUT_LOBJREF( frame, 0, objref );

			/* Commit */
			fp->prev = frame;
			stp->fp = frame;

#ifdef _DEBUG
			if( method_debug )
			{
				char buf[255], cbuf[255];
				sprintMethodInfo( buf, mi );
				sprintClassName( cbuf, obj->cf, obj->cf->this_class );
				printf( "Java method %s (from class %s)\n", buf,cbuf );
			}
#endif /* _DEBUG */
			
			if( callstack )
				(*callstack)( frame, 1 /* push */ );
			
			return 0;
		}
	}

	return newCompleteException( stp, InternalError, "Bad MethodRef" );
}


/******************************************************************************
 * M
 *****************************************************************************/

void monitorenter( Frame* fp )
{
	ObjRef* objref;

	INSTRUCTION( "%p: monitorenter\n", fp->pc );

	POP_OBJREF( fp, objref );
}


void monitorexit( Frame* fp )
{
	ObjRef* objref;

	INSTRUCTION( "%p: monitorexit\n", fp->pc );

	POP_OBJREF( fp, objref );
}


int multianewarray( Stack* stp, u2 index, u1 dimension )
{
	Frame* fp = stp->fp;
	ObjRef* objref;
	int32 length[10];
	u1 basetype = T_LAST;
	u2 size = 0;
	int n;

	INSTRUCTION( "%p: multianewarray %d, %d\n", fp->pc, index, dimension );

	if( (dimension < 1) || (dimension > 10) )
	{
		char buf[255];
		sprintf( buf, "multianewarray with dimension = %d not implemented\n", dimension );
		return newCompleteException( stp, InternalError, buf );
	}

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Class ) )
	{
		CONSTANT_Class_info* cp = (CONSTANT_Class_info*) fp->cf->constant_pool[index];

		if( CHECKCONSTANT( fp->cf, cp->name_index, CONSTANT_Utf8 ) )
		{
			CONSTANT_Utf8_info* name = (CONSTANT_Utf8_info*) fp->cf->constant_pool[cp->name_index];

			for( n = 0; name->bytes[n] == '['; n++ );

			if( n == 0 )
				return newCompleteException( stp, InternalError, "multianewarray called for non-array\n" );

			switch( name->bytes[n] )
			{
			case 'B': basetype = T_BYTE;    size = 1; break;
			case 'C': basetype = T_CHAR;    size = 1; break; /* will eventually be 2 */
			case 'D': basetype = T_DOUBLE;  size = 8; break;
			case 'F': basetype = T_FLOAT;   size = 4; break;
			case 'I': basetype = T_INT;     size = 4; break;
			case 'J': basetype = T_LONG;    size = 8; break;
			case 'L': basetype = T_OBJREF;  size = sizeof( ObjRef*); break;
			case 'S': basetype = T_SHORT;   size = 2; break;
			case 'Z': basetype = T_BOOLEAN; size = 1; break;
			}

			for( n = 0; n < dimension; n++ )
			{
				POP_INT( fp, length[dimension-n-1] );

				if( length[dimension-n-1] < 0 )
				{
					char buf[256];
					sprintf( buf, "%d", (int) length[dimension-n-1] );
					return newCompleteException( stp, NegativeArraySizeException, buf );
				}
			}

			objref = createMultiObjrefArray( dimension, length, size, basetype );
			PUSH_OBJREF( fp, objref );

			/* add to array pool */
			addToList( objref, &arrayPool );
			return 0;
		}
	}

	return newCompleteException( stp, InternalError, "could not create multianewarray" );
}


/******************************************************************************
 * N
 *****************************************************************************/

int newarray( Stack* stp, u1 type )
{
	Frame* fp = stp->fp;
	ObjRef* objref;
	u4 size = 0;
	int32 length;

	INSTRUCTION( "%p: newarray %d\n", fp->pc, type );

	POP_INT( fp, length );

	if( length < 0 )
	{
		char buf[256];
		sprintf( buf, "%d", (int) length );
		return newCompleteException( stp, NegativeArraySizeException, buf );
	}

	switch( type )
	{
	case T_BOOLEAN: size = 1; break;
	case T_CHAR:	size = 1; break;
	case T_FLOAT:	size = 4; break;
	case T_DOUBLE:	size = 8; break;
	case T_BYTE:	size = 1; break;
	case T_SHORT:	size = 2; break;
	case T_INT:		size = 4; break;
	case T_LONG:	size = 8; break;
	default: printf( "ERROR: bad array type\n" ); break;
	}

	objref = newObjectRefAndArray( type, length, size);
	PUSH_OBJREF( fp, objref );

	/* store in array pool */
	addToList( objref, &arrayPool );
	return 0;
}


int newobject( Stack* stp, u2 index )
{
	Frame *fp = stp->fp;
	ObjRef* objref;

	INSTRUCTION( "%p: new %d\n", fp->pc, index );

	if( CHECKCONSTANT( fp->cf, index, CONSTANT_Class ) )
	{
		CONSTANT_Class_info* cp = (CONSTANT_Class_info*) fp->cf->constant_pool[index];

		if( !cp->cf )
			linkClass( fp->cf, index, stp );

		if( !cp->cf )
			return stp->retcode; /* linkClass looks after exception generation */

		objref = newObjectRefAndObject( cp->cf );
		PUSH_OBJREF( fp, objref );

		/* store in object pool */
		addToList( objref, &objectPool );
		return 0;
	}

	return newCompleteException( stp, InternalError, "Failed to lookup CONSTANT_Class" );
}


/******************************************************************************
 * P
 *****************************************************************************/

void pop( Frame* fp )
{
	u4 dummy;

	INSTRUCTION( "%p: pop\n", fp->pc );

	POP_ANY( fp, dummy );
	*(fp->sp) = 0; /* clear stack entry */
	*(fp->tsp) = 0; /* clear type stack entry */
}


void pop2( Frame* fp )
{
	u4 dummy[2];

	INSTRUCTION( "%p: pop2\n", fp->pc );

	POP_ANY( fp, dummy[0] );
	*(fp->sp) = 0; /* clear stack entry */
	*(fp->tsp) = 0; /* clear type stack entry */

	POP_ANY( fp, dummy[1] );
	*(fp->sp) = 0; /* clear stack entry */
	*(fp->tsp) = 0; /* clear type stack entry */
}


/******************************************************************************
 * R
 *****************************************************************************/

void ret( Frame* fp, u2 index )
{
	ObjRef* obj;

	INSTRUCTION( "%p: ret %d\n", fp->pc, index );

	GET_LRETADDR( fp, index, obj );
	fp->pc = (u1*) obj;
}


/******************************************************************************
 * S
 *****************************************************************************/

#include "short.c"


/******************************************************************************
 * T
 *****************************************************************************/

void tableswitch( Frame* fp )
{
	int32 index, def, low, high, offset;
	u4 pcoff = fp->pc - fp->min_pc;
	u1* pc;

	pcoff = ((pcoff)/4)*4 + 3;
	pc = fp->min_pc + pcoff;

	def = (int32) PEEKU4( pc ); pc += 4;
	low = (int32) PEEKU4( pc ); pc += 4;
	high = (int32) PEEKU4( pc ); pc += 4;

	INSTRUCTION( "%p: tableswitch %d %d %d\n", fp->pc, (int) def, (int) low, (int) high );

	POP_INT( fp, index );

	if( (index < low) || (index > high) )
	{
		fp->pc += def;
		return;
	}

	pc += (index - low) * 4;
	offset = (int32) PEEKU4( pc );

	fp->pc += offset;
}


/******************************************************************************
 * V
 *****************************************************************************/

int vreturn( Stack* stp )
{
	Frame* fp = stp->fp;
	int stopOnPop = stp->fp->stopOnPop;

	INSTRUCTION( "%p: return\n", fp->pc );
	stp->fp = fp->next;
	freeStackFrame( fp, fp->mi );
	fp = stp->fp;

	if( fp )
		fp->prev = NULL;

#ifdef _DEBUG
	if( fp && method_debug )
	{
		char buf[255];
		sprintMethodInfo( buf, fp->mi );
		printf( "Returned to method %s\n", buf );
	}
#endif /* _DEBUG */

	if( callstack )
		(*callstack)( fp, 0 /* pop */ );

	return ( (!fp || stopOnPop) ? 1 : 0 );
}


/******************************************************************************
 * W
 *****************************************************************************/

int wide( Stack* stp, u1 inst )
{
	Frame* fp = stp->fp;
	int size = 0;

	switch( inst )
	{
	case 0x15: iload_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x16: lload_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x17: fload_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x18: dload_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x19: aload_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x36: istore_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x37: lstore_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x38: fstore_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x39: dstore_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x3a: astore_n( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	case 0x84: iinc_w( fp, PEEKU2(fp->pc+1), PEEKU2(fp->pc+3) ); size = 6; break;
	case 0xa9: ret( fp, PEEKU2(fp->pc+1) ); size = 4; break;
	default:
		{
			char buf[255];
#ifdef _DEBUG
			sprintf( buf, "unknown wide instruction 0x%x (%s)", inst, instructionTable[inst].name );
#else
			sprintf( buf, "unknown wide instruction 0x%x", inst );
#endif
			printf( "INTERNAL ERROR: %s\n", buf );
			return newCompleteException( stp, InternalError, buf );
		}
	}

	fp->pc += size; /* the wide part */
	return 0;
}

/************************* End of Instructions *******************************/

int exceptionHandler( Stack* stp )
{
	Frame* fp = stp->fp;

#ifdef _DEBUG
	if( verbose )
	{
		if( fp && fp->mi )
		{
			char buf[255];
			sprintMethodInfo( buf, fp->mi );
			printf( "exception handler: %s\n", buf );
		}
		else
			printf( "exception handler: no method\n" );
	}
#endif /* _DEBUG */

	if( fp && fp->ex && (fp->ex->kind == JavaException) && fp->ex->ex )
	{
		/* check exception table for current method */
		if( fp->mi && fp->mi->ca && fp->mi->ca->exception_table )
		{
			Exception_table** exception_table = fp->mi->ca->exception_table;
			u2 exception_table_length = fp->mi->ca->exception_table_length;
			u4 pc = fp->pc - fp->min_pc;
			u2 n;

			for( n = 0; n < exception_table_length; n++ )
			{
				if( (pc >= exception_table[n]->start_pc)
					&& (pc <= exception_table[n]->end_pc) )
				{
					if( CHECKCONSTANT( fp->cf, exception_table[n]->catch_type, CONSTANT_Class) )
					{
						const ClassFile* catch_type = ((CONSTANT_Class_info*) fp->cf->constant_pool[exception_table[n]->catch_type])->cf;
						const ClassFile* ex_type = fp->ex->ex->cf;

						if( !catch_type )
						{
							linkClass( fp->cf, exception_table[n]->catch_type, stp );
							catch_type = ((CONSTANT_Class_info*) fp->cf->constant_pool[exception_table[n]->catch_type])->cf;
							if( !catch_type )
								break;
						}

						if( subclassOf( ex_type, catch_type ) )
						{
							/* clear operand stack as per spec. */
							fp->sp  = &fp->local[fp->mi->ca->max_locals]; /* end of locals */
							fp->tsp = &fp->tlocal[fp->mi->ca->max_locals]; /* end of tlocals */

							fp->pc = fp->min_pc + exception_table[n]->handler_pc;
							PUSH_OBJREF( fp, fp->ex->ex );
							fp->ex = NULL;
							return 0; /* Exception handled */
						}
					}
				}
			}
		}

		/* pop stack frame and propogate the exception */
		stp->fp = fp->next;

		if( stp->fp )
			stp->fp->ex = fp->ex;
		else
			stp->ex = fp->ex;

		freeStackFrame( fp, fp->mi );
		fp = stp->fp;

		if( fp )
			fp->prev = NULL;

		if( callstack )
			(*callstack)( fp, 0 /* pop */ );

		return (fp == NULL ? 2 : 0 );
	}

	return 2;
}


int executeInstruction( Stack* stp )
{
	Frame* fp = stp->fp;
	int retcode = 0;
	u1 inst;
	u4 start_pc;

	halt = 0;

	if( !fp )
	{
		newCompleteException( stp, InternalError, "Stack empty" );
		return 2; /* non-trappable error */
	}

	/* exception handler - propogate the current or a newly initialised exception */
	if( fp->ex )
		return exceptionHandler( stp );

	inst = *(fp->pc);
	start_pc = fp->pc - fp->min_pc;

	if( instruction )
		(*instruction)( fp );

	instruction_count++;

#ifdef _DEBUG
	instructionTable[inst].count++;
#endif

	switch( inst )
	{
	case 0x00: /* nop */; break;
	case 0x01: aconst_null( fp ); break;
	case 0x02: iconst_n( fp, -1 ); break;
	case 0x03: iconst_n( fp, 0 ); break;
	case 0x04: iconst_n( fp, 1 ); break;
	case 0x05: iconst_n( fp, 2 ); break;
	case 0x06: iconst_n( fp, 3 ); break;
	case 0x07: iconst_n( fp, 4 ); break;
	case 0x08: iconst_n( fp, 5 ); break;
	case 0x09: lconst_n( fp, 0 ); break;
	case 0x0a: lconst_n( fp, 1 ); break;
	case 0x0b: fconst_n( fp, 0 ); break;
	case 0x0c: fconst_n( fp, 1 ); break;
	case 0x0d: fconst_n( fp, 2 ); break;
	case 0x0e: dconst_n( fp, 0 ); break;
	case 0x0f: dconst_n( fp, 1 ); break;
	case 0x10: bipush( fp, PEEKU1(fp->pc) ); break;
	case 0x11: sipush( fp, PEEKU2(fp->pc) ); break;
	case 0x12: retcode = ldc( stp, PEEKU1(fp->pc) ); break;
	case 0x13: retcode = ldc( stp, PEEKU2(fp->pc) ); break;
	case 0x14: retcode = ldc2_w( stp, PEEKU2(fp->pc) ); break;
	case 0x15: iload_n( fp, PEEKU1(fp->pc) ); break;
	case 0x16: lload_n( fp, PEEKU1(fp->pc) ); break;
	case 0x17: fload_n( fp, PEEKU1(fp->pc) ); break;
	case 0x18: dload_n( fp, PEEKU1(fp->pc) ); break;
	case 0x19: aload_n( fp, PEEKU1(fp->pc) ); break;
	case 0x1a: iload_n( fp, 0 ); break; /* iload_0 */
	case 0x1b: iload_n( fp, 1 ); break; /* iload_1 */
	case 0x1c: iload_n( fp, 2 ); break; /* iload_2 */
	case 0x1d: iload_n( fp, 3 ); break; /* iload_3 */
	case 0x1e: lload_n( fp, 0 ); break; /* lload_0 */
	case 0x1f: lload_n( fp, 1 ); break; /* lload_1 */
	case 0x20: lload_n( fp, 2 ); break; /* lload_2 */
	case 0x21: lload_n( fp, 3 ); break; /* lload_3 */
	case 0x22: fload_n( fp, 0 ); break; /* fload_0 */
	case 0x23: fload_n( fp, 1 ); break; /* fload_1 */
	case 0x24: fload_n( fp, 2 ); break; /* fload_2 */
	case 0x25: fload_n( fp, 3 ); break; /* fload_3 */
	case 0x26: dload_n( fp, 0 ); break; /* dload_0 */
	case 0x27: dload_n( fp, 1 ); break; /* dload_1 */
	case 0x28: dload_n( fp, 2 ); break; /* dload_2 */
	case 0x29: dload_n( fp, 3 ); break; /* dload_3 */
	case 0x2a: aload_n( fp, 0 ); break; /* aload_0 */
	case 0x2b: aload_n( fp, 1 ); break; /* aload_1 */
	case 0x2c: aload_n( fp, 2 ); break; /* aload_2 */
	case 0x2d: aload_n( fp, 3 ); break; /* aload_3 */
	case 0x2e: retcode = iaload( stp ); break;
	case 0x2f: retcode = laload( stp ); break;
	case 0x30: retcode = faload( stp ); break;
	case 0x31: retcode = daload( stp ); break;
	case 0x32: retcode = aaload( stp ); break;
	case 0x33: retcode = baload( stp ); break;
	case 0x34: retcode = caload( stp ); break;
	case 0x35: retcode = saload( stp ); break;
	case 0x36: istore_n( fp, PEEKU1(fp->pc) ); break; /* istore */
	case 0x37: lstore_n( fp, PEEKU1(fp->pc) ); break; /* lstore */
	case 0x38: fstore_n( fp, PEEKU1(fp->pc) ); break; /* fstore */
	case 0x39: dstore_n( fp, PEEKU1(fp->pc) ); break; /* dstore */
	case 0x3a: astore_n( fp, PEEKU1(fp->pc) ); break; /* astore */
	case 0x3b: istore_n( fp, 0 ); break; /* istore_0 */
	case 0x3c: istore_n( fp, 1 ); break; /* istore_1 */
	case 0x3d: istore_n( fp, 2 ); break; /* istore_2 */
	case 0x3e: istore_n( fp, 3 ); break; /* istore_3 */
	case 0x3f: lstore_n( fp, 0 ); break; /* lstore_0 */
	case 0x40: lstore_n( fp, 1 ); break; /* lstore_1 */
	case 0x41: lstore_n( fp, 2 ); break; /* lstore_2 */
	case 0x42: lstore_n( fp, 3 ); break; /* lstore_3 */
	case 0x43: fstore_n( fp, 0 ); break; /* fstore_0 */
	case 0x44: fstore_n( fp, 1 ); break; /* fstore_1 */
	case 0x45: fstore_n( fp, 2 ); break; /* fstore_2 */
	case 0x46: fstore_n( fp, 3 ); break; /* fstore_3 */
	case 0x47: dstore_n( fp, 0 ); break; /* dstore_0 */
	case 0x48: dstore_n( fp, 1 ); break; /* dstore_1 */
	case 0x49: dstore_n( fp, 2 ); break; /* dstore_2 */
	case 0x4a: dstore_n( fp, 3 ); break; /* dstore_3 */
	case 0x4b: astore_n( fp, 0 ); break; /* astore_0 */
	case 0x4c: astore_n( fp, 1 ); break; /* astore_1 */
	case 0x4d: astore_n( fp, 2 ); break; /* astore_2 */
	case 0x4e: astore_n( fp, 3 ); break; /* astore_3 */
	case 0x4f: retcode = iastore( stp ); break;
	case 0x50: retcode = lastore( stp ); break;
	case 0x51: retcode = fastore( stp ); break;
	case 0x52: retcode = dastore( stp ); break;
	case 0x53: retcode = aastore( stp ); break;
	case 0x54: retcode = bastore( stp ); break;
	case 0x55: retcode = castore( stp ); break;
	case 0x56: retcode = sastore( stp ); break;
	case 0x57: pop( fp ); break;
	case 0x58: pop2( fp ); break;
	case 0x59: dup( fp ); break;
	case 0x5a: dup_x1( fp ); break;
	case 0x5b: dup_x2( fp ); break;
	case 0x5c: dup2( fp ); break;
	case 0x5d: dup2_x1( fp ); break;
	case 0x5e: dup2_x2( fp ); break;
	case 0x5f: swap( fp ); break;
	case 0x60: iadd( fp ); break;
	case 0x61: ladd( fp ); break;
	case 0x62: fadd( fp ); break;
	case 0x63: dadd( fp ); break;
	case 0x64: isub( fp ); break;
	case 0x65: lsub( fp ); break;
	case 0x66: fsub( fp ); break;
	case 0x67: dsub( fp ); break;
	case 0x68: imul( fp ); break;
	case 0x69: lmul( fp ); break;
	case 0x6a: fmul( fp ); break;
	case 0x6b: dmul( fp ); break;
	case 0x6c: retcode = idiv( stp ); break;
	case 0x6d: retcode = ldivide( stp ); break;
	case 0x6e: fdiv( fp ); break;
	case 0x6f: ddiv( fp ); break;
	case 0x70: retcode = irem( stp ); break;
	case 0x71: retcode = lrem( stp ); break;
	case 0x72: frem( fp ); break;
	case 0x73: dremainder( fp ); break;
	case 0x74: ineg( fp ); break;
	case 0x75: lneg( fp ); break;
	case 0x76: fneg( fp ); break;
	case 0x77: dneg( fp ); break;
	case 0x78: ishl( fp ); break;
	case 0x79: lshl( fp ); break;
	case 0x7a: ishr( fp ); break;
	case 0x7b: lshr( fp ); break;
	case 0x7c: iushr( fp ); break;
	case 0x7d: lushr( fp ); break;
	case 0x7e: iand( fp ); break;
	case 0x7f: land( fp ); break;
	case 0x80: ior( fp ); break;
	case 0x81: lor( fp ); break;
	case 0x82: ixor( fp ); break;
	case 0x83: lxor( fp ); break;
	case 0x84: iinc( fp, PEEKU1(fp->pc), PEEKU1(fp->pc + 1) ); break;
	case 0x85: i2l( fp ); break;
	case 0x86: i2f( fp ); break;
	case 0x87: i2d( fp ); break;
	case 0x88: l2i( fp ); break;
	case 0x89: l2f( fp ); break;
	case 0x8a: l2d( fp ); break;
	case 0x8b: f2i( fp ); break;
	case 0x8c: f2l( fp ); break;
	case 0x8d: f2d( fp ); break;
	case 0x8e: d2i( fp ); break; /* ### was 0x83 */
	case 0x8f: d2l( fp ); break;
	case 0x90: d2f( fp ); break;
	case 0x91: i2b( fp ); break;
	case 0x92: i2c( fp ); break;
	case 0x93: i2s( fp ); break;
	case 0x94: lcmp( fp ); break;
	case 0x95: fcmpl( fp ); break;
	case 0x96: fcmpg( fp ); break;
	case 0x97: dcmpl( fp ); break;
	case 0x98: dcmpg( fp ); break;
	case 0x99: ifeq( fp, PEEKU2(fp->pc) ); break;
	case 0x9a: ifne( fp, PEEKU2(fp->pc) ); break;
	case 0x9b: iflt( fp, PEEKU2(fp->pc) ); break;
	case 0x9c: ifge( fp, PEEKU2(fp->pc) ); break;
	case 0x9d: ifgt( fp, PEEKU2(fp->pc) ); break;
	case 0x9e: ifle( fp, PEEKU2(fp->pc) ); break;
	case 0x9f: if_icmpeq( fp, PEEKU2(fp->pc) ); break;
	case 0xa0: if_icmpne( fp, PEEKU2(fp->pc) ); break;
	case 0xa1: if_icmplt( fp, PEEKU2(fp->pc) ); break;
	case 0xa2: if_icmpge( fp, PEEKU2(fp->pc) ); break;
	case 0xa3: if_icmpgt( fp, PEEKU2(fp->pc) ); break;
	case 0xa4: if_icmple( fp, PEEKU2(fp->pc) ); break;
	case 0xa5: if_acmpeq( fp, PEEKU2(fp->pc) ); break;
	case 0xa6: if_acmpne( fp, PEEKU2(fp->pc) ); break;
	case 0xa7: goto_inst( fp, PEEKU2(fp->pc) ); break;
	case 0xa8: jsr( fp, PEEKU2(fp->pc) ); break;
	case 0xa9: ret( fp, PEEKU1(fp->pc) ); break;
	case 0xaa: tableswitch( fp ); break;
	case 0xab: lookupswitch( fp ); break;
	case 0xac: return ireturn( stp );
	case 0xad: return lreturn( stp );
	case 0xae: return freturn( stp );
	case 0xaf: return dreturn( stp );
	case 0xb0: return areturn( stp );
	case 0xb1: return vreturn( stp );
	case 0xb2: retcode = getstatic( stp, PEEKU2(fp->pc) ); break;
	case 0xb3: retcode = putstatic( stp, PEEKU2(fp->pc) ); break;
	case 0xb4: retcode = getfield( stp, PEEKU2(fp->pc) ); break;
	case 0xb5: retcode = putfield( stp, PEEKU2(fp->pc) ); break;
	case 0xb6: retcode = invokevirtual( stp, PEEKU2(fp->pc) ); break;
	case 0xb7: retcode = invokespecial( stp, PEEKU2(fp->pc) ); break;
	case 0xb8: retcode = invokestatic( stp, PEEKU2(fp->pc) ); break;
	case 0xb9: retcode = invokeinterface( stp, PEEKU2(fp->pc), PEEKU1(fp->pc+2) ); break;
	case 0xbb: retcode = newobject( stp, PEEKU2(fp->pc) ); break;
	case 0xbc: retcode = newarray( stp, PEEKU1(fp->pc) ); break;
	case 0xbd: retcode = anewarray( stp, PEEKU2(fp->pc) ); break;
	case 0xbe: retcode = arraylength( stp ); break;
	case 0xbf: retcode = athrow( stp ); break;
	case 0xc0: retcode = checkcast( stp, PEEKU2(fp->pc) ); break;
	case 0xc1: retcode = instanceof( stp, PEEKU2(fp->pc) ); break;
	case 0xc2: monitorenter( fp ); break;
	case 0xc3: monitorexit( fp ); break;
	case 0xc4: retcode = wide( stp, PEEKU1(fp->pc) ); break;
	case 0xc5: retcode = multianewarray( stp, PEEKU2(fp->pc), PEEKU1(fp->pc+2) ); break;
	case 0xc6: ifnull( fp, PEEKU2(fp->pc) ); break;
	case 0xc7: ifnonnull( fp, PEEKU2(fp->pc) ); break;
	case 0xc8: goto_w( fp, PEEKU4(fp->pc) ); break;
	case 0xc9: jsr_w( fp, PEEKU4(fp->pc) ); break;
	default:
		{
			char buf[255];
#ifdef _DEBUG
			sprintf( buf, "unknown instruction 0x%x (%s)", inst, instructionTable[inst].name );
#else
			sprintf( buf, "unknown instruction 0x%x", inst );
#endif
			printf( "INTERNAL ERROR: %s\n", buf );
			return newCompleteException( stp, InternalError, buf );
		}
	}

#ifdef _DEBUG
	if( fp->sp > (u4*) fp->tlocal )
	{
		printf( "INTERNAL ERROR: stack overrun (%p > %p)\n", fp->sp, fp->tlocal );
		halt = 1;
		return newCompleteException( stp, InternalError, "Stack overrun" );
	}

	if( fp->sp < &fp->local[fp->mi->ca->max_locals] )
	{
		printf( "INTERNAL ERROR: stack underrun (%p < %p)\n", fp->sp, fp->tlocal );
		halt = 1;
		return newCompleteException( stp, InternalError, "Stack underrun" );
	}
#endif

	if( halt )
	{
#ifdef _DEBUG
		printf( "INTERNAL ERROR: Stack type check failed with instruction 0x%x (%s)\n", inst, instructionTable[inst].name );
#else
		printf( "INTERNAL ERROR: Stack type check failed with instruction 0x%x\n", inst );
#endif
		return newCompleteException( stp, InternalError, "Stack type check failed" );
	}

	/* sanity check pc otherwise exception handle may fail */
	if( (fp->pc + instructionTable[inst].incr) >= fp->max_pc )
	{
		printf( "INTERNAL ERROR: PC exceeds code boundary\n" );
		return newCompleteException( stp, InternalError, "PC exceeds code boundary" );
	}

	/* Exception handler - propogate a new exception */
	if( (stp->fp && stp->fp->ex) || stp->ex )
		return exceptionHandler( stp );

	/* increment the program counter but only if a new uninitialised exception
	 * has not been thrown */
	if( !(fp->ex) && instructionTable[inst].incr )
	{
		u4 pc = fp->pc - fp->min_pc;
		if( pc != start_pc )
		{
			newCompleteException( stp, InternalError, "PC out of sync" );
			return 2; /* non-trapable error */
		}

		fp->pc += instructionTable[inst].incr;
	}

	return retcode;
}


int newCompleteException( Stack* stp, int kind, const char* mess )
{
	Frame* fp = stp->fp;
	ClassFile* excf;
	const char* exclass = NULL;

	if( !fp )
	{
		stp->ex = (Exception*) javavm_malloc( sizeof( Exception ) );
		stp->ex->kind = kind;
		stp->ex->message = (char*) mess;
		stp->ex->ex = NULL;
		return 2;
	}

	exclass = exceptionKindToString( kind );

	if( verbose )
	{
		printf( "JavaException: %s: %s\n", exclass, (mess ? mess : "") );
		whereAreWe( fp );
	}

	if( exclass )
	{
		if( (excf = getClass( stp, exclass )) )
		{
			int retcode;
		
			fp->ex = (Exception*) javavm_malloc( sizeof( Exception ) );
			fp->ex->kind = JavaException;
			fp->ex->message = NULL;
			fp->ex->ex = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
			fp->ex->ex->cf = excf;
			fp->ex->ex->type = 0;
			fp->ex->ex->data = (u4*) javavm_malloc( 4 * excf->object_size );
			fp->ex->ex->length = 1;
			memset( fp->ex->ex->data, 0, 4 * excf->object_size );

			/* reset stack pointer to clear stack */
			if( fp->mi && fp->mi->ca )
			{
				fp->sp = fp->local + fp->mi->ca->max_locals;
				fp->tsp = fp->tlocal + fp->mi->ca->max_locals;
			}

			if( mess )
			{
				ObjRef* str = newJavaLangString( (u1*) mess, strlen(mess) );
				PUSH_OBJREF( fp, str );
				retcode = forceExecuteJavaMethod( fp->ex->ex, stp, "<init>", "(Ljava/lang/String;)V" );
			}
			else
				retcode = forceExecuteJavaMethod( fp->ex->ex, stp, "<init>", "()V" );

			return retcode;
		}
	}

	printf( "WARNING: Could not load JavaException class \"%s\"\n", exclass );

	fp->ex = (Exception*) javavm_malloc( sizeof( Exception ) );
	fp->ex->kind = kind;
	fp->ex->message = (char*) mess;
	fp->ex->ex = NULL;

	return 2;
}


int lookupJavaMethod( ClassFile* cf, Stack* stp, const char* name, const char* desc )
{
	char buf[255];
	u2 n;

	for( n = 0; n < cf->methods_count; n++ )
	{
		if( cf->methods[n]
			&& checkSimpleName( cf, cf->methods[n]->name_index, name )
			&& checkSimpleName( cf, cf->methods[n]->descriptor_index, desc ) )
		{
			break;
		}
	}

	if( n >= cf->methods_count )
	{
		char buf[255]; char cname[255];
		sprintClassName( cname, cf, cf->this_class );
		sprintf( buf, "No \"%s\" method for class \"%s\"", name, cname );
		newCompleteException( stp, NoSuchMethodError, buf );
		return -1; /* exception */
	}

	if( (cf->methods[n]->access_flags & ACC_NATIVE)
		|| !cf->methods[n]->ca || !cf->methods[n]->ca->code )
	{
		char ex[255];
		sprintMethodInfo( buf, cf->methods[n] );
		sprintf( ex, "No Java code for method \"%s\"\n", buf );
		printf( "%s", ex );
		newCompleteException( stp, NoSuchMethodError, ex );
		return -1; /* exception */
	}

	return n;
}


int forceExecuteJavaMethod( ObjRef* objref, Stack* stp, const char* name, const char* desc )
{
	Frame *fp = stp->fp;
	u4 value;
	u1 type;
	int n;

	n = lookupJavaMethod( objref->cf, stp, name, desc );
	if( n == -1 ) return 2; /* exception */

#ifdef _DEBUG
	if( verbose & 2 )
	{
		char buf[255];
		sprintClassName( buf, objref->cf, objref->cf->this_class );
		printf( "Forcefully constructing class \"%s\" (ordinary method)...\n", buf );
	}
#endif /* _DEBUG */

	stp->fp = newStackFrame( objref->cf->methods[n], fp );

	PUT_LOBJREF( stp->fp, 0, objref ); /* this */

	/* pop args */
	for( n = objref->cf->methods[n]->args; n > 0; n-- )
	{
		POP_ANY_TYPE( fp, value, type );
		PUT_LANY_TYPE( stp->fp, n, value, type );
	}

	/* Commit */
	if( fp )
		fp->prev = stp->fp;

	if( callstack )
		(*callstack)( stp->fp, 1 /* push */ );

#ifdef _DEBUG
	if( method_debug )
	{
		char buf[255], cbuf[255];
		sprintMethodInfo( buf, objref->cf->methods[n] );
		sprintClassName( cbuf, objref->cf, objref->cf->this_class );
		printf( "Java method %s (from class %s)\n", buf,cbuf );
	}
#endif /* _DEBUG */

	return 0;
}


int forceExecuteJavaMethodWithArgs( ObjRef* objref, Stack* stp, const char* name, const char* desc, int argc, u4 argv[], u1 argvt[] )
{
	Frame *fp = stp->fp;
	int n;

		n = lookupJavaMethod( objref->cf, stp, name, desc );
		if( n == -1 ) return 2; /* exception */

#ifdef _DEBUG
	if( verbose & 2 )
	{
		char buf[255];
		sprintClassName( buf, objref->cf, objref->cf->this_class );
		printf( "Forcefully constructing class \"%s\" (ordinary method)...\n", buf );
	}
#endif /* _DEBUG */

	stp->fp = newStackFrame( objref->cf->methods[n], fp );

	PUT_LOBJREF( stp->fp, 0, objref ); /* this */

	/* store args */
	if( argc && argv )
	{
		int n;

		for( n = 0; n < argc; n++ )
		{
			PUT_LANY_TYPE( stp->fp, n+1, argv[n], argvt[n] );
		}
	}

	/* Commit */
	if( fp )
		fp->prev = stp->fp;

	if( callstack )
		(*callstack)( stp->fp, 1 /* push */ );

#ifdef _DEBUG
	if( method_debug )
	{
		char buf[255], cbuf[255];
		sprintMethodInfo( buf, objref->cf->methods[n] );
		sprintClassName( cbuf, objref->cf, objref->cf->this_class );
		printf( "Java method %s (from class %s)\n", buf,cbuf );
	}
#endif /* _DEBUG */

	return 0;
}


int forceExecuteVirtualJavaMethod( ObjRef* objref, Stack* stp, const char* name, const char* desc, int stopOnPop )
{
	Frame *fp = stp->fp;
	char buf[255];
	u4 value;
	u1 type;
	int n;

	for( n = 0; n < objref->cf->virtuals_count; n++ )
	{
		if( objref->cf->virtuals[n]
			&& checkSimpleName( objref->cf->virtuals[n]->cf, objref->cf->virtuals[n]->name_index, name )
			&& checkSimpleName( objref->cf->virtuals[n]->cf, objref->cf->virtuals[n]->descriptor_index, desc ) )
		{
			break;
		}
	}

	if( n >= objref->cf->virtuals_count )
	{
		char buf[255]; char cname[255];
		sprintClassName( cname, objref->cf, objref->cf->this_class );
		sprintf( buf, "No \"%s\" method for class \"%s\"", name, cname );
		return newCompleteException( stp, NoSuchMethodError, buf );
	}

	if( (objref->cf->virtuals[n]->access_flags & ACC_NATIVE) || !objref->cf->virtuals[n]->ca || !objref->cf->virtuals[n]->ca->code )
	{
		char ex[255];
		sprintMethodInfo( buf, objref->cf->virtuals[n] );
		sprintf( ex, "No Java code for method \"%s\"\n", buf );
		printf( "%s", ex );
		return newCompleteException( stp, NoSuchMethodError, ex );
	}

#ifdef _DEBUG
	if( verbose & 2 )
	{
		sprintClassName( buf, objref->cf, objref->cf->this_class );
		printf( "Forcefully constructing class \"%s\" (virtual method)...\n", buf );
	}
#endif /* _DEBUG */

	stp->fp = newStackFrame( objref->cf->virtuals[n], fp );
	stp->fp->stopOnPop = stopOnPop;

	PUT_LOBJREF( stp->fp, 0, objref ); /* this */

	/* pop args */
	for( n = objref->cf->virtuals[n]->args; n > 0; n-- )
	{
		POP_ANY_TYPE( fp, value, type );
		PUT_LANY_TYPE( stp->fp, n, value, type );
	}

	/* Commit */
	if( fp )
		fp->prev = stp->fp;

	if( callstack )
		(*callstack)( stp->fp, 1 /* push */ );

#ifdef _DEBUG
	if( method_debug )
	{
		char buf[255], cbuf[255];
		sprintMethodInfo( buf, objref->cf->methods[n] );
		sprintClassName( cbuf, objref->cf, objref->cf->this_class );
		printf( "Java method %s (from class %s)\n", buf,cbuf );
	}
#endif /* _DEBUG */

	return 0;
}


int forceExecuteStaticJavaMethod( ClassFile* cf, Stack* stp, const char* name, const char* desc )
{
	Frame *fp = stp->fp;
	char buf[255];
	u4 value;
	u1 type;
	int n;

	n = lookupJavaMethod( cf, stp, name, desc );
	if( n == -1 ) return 2; /* exception */

	if( !(cf->methods[n]->access_flags & ACC_STATIC) )
	{
		char ex[255];
		sprintMethodInfo( buf, cf->methods[n] );
		sprintf( ex, "Method not static \"%s\"\n", buf );
		printf( "%s", ex );
		return newCompleteException( stp, NoSuchMethodError, ex );
	}

#ifdef _DEBUG
	if( verbose & 2 )
	{
		sprintClassName( buf, cf, cf->this_class );
		printf( "Forcefully constructing class \"%s\" (static method)...\n", buf );
	}
#endif /* _DEBUG */

	stp->fp = newStackFrame( cf->methods[n], fp );

	/* pop args */
	for( n = cf->methods[n]->args; n > 0; n-- )
	{
		POP_ANY_TYPE( fp, value, type );
		PUT_LANY_TYPE( stp->fp, n-1, value, type );
	}

	/* Commit */
	if( fp )
		fp->prev = stp->fp;

	if( callstack )
		(*callstack)( stp->fp, 1 /* push */ );

#ifdef _DEBUG
	if( method_debug )
	{
		char buf[255], cbuf[255];
		sprintMethodInfo( buf, cf->methods[n] );
		sprintClassName( cbuf, cf, cf->this_class );
		printf( "Java method %s (from class %s)\n", buf,cbuf );
	}
#endif /* _DEBUG */

	return 0;
}


int executeCLInit( ClassFile* cf, Stack* stp )
{
	Frame *fp = stp->fp;
	int retcode = 0;
	char buf[255];
	u2 n;

	for( n = 0; n < cf->methods_count; n++ )
	{
		if( cf->methods[n] && checkSimpleName( cf, cf->methods[n]->name_index, "<clinit>" ) )
			break;
	}

	if( n >= cf->methods_count )
	{
		/* printf( "No \"<clinit>\" method for class \"" ); writeClassName( cf, cf->this_class ); printf( "\"\n" ); */
		return 1 /* exit */;
	}

	if( (cf->methods[n]->access_flags & ACC_NATIVE) || !cf->methods[n]->ca || !cf->methods[n]->ca->code )
	{
		char ex[255];
		sprintClassName( buf, cf, cf->this_class );
		sprintf( ex, "No Java code for method \"%s.<clinit>\"\n", buf );
		printf( "%s", ex );
		return newCompleteException( stp, NoSuchMethodError, ex );
	}

#ifdef _DEBUG
	if( verbose & 2 )
	{
		sprintClassName( buf, cf, cf->this_class );
		printf( "Initialising class \"%s\"...\n", buf );
	}
#endif

	stp->fp = newStackFrame( cf->methods[n], NULL );

	/* next would result in execution - must stop after <clinit> */
	if( stp->fp )
		stp->fp->init = fp;

#ifdef _DEBUG
	if( method_debug )
	{
		char buf[255];
		sprintMethodInfo( buf, cf->methods[n] );
		printf( "Java method %s\n", buf );
	}
#endif /* _DEBUG */

	if( callstack )
		(*callstack)( stp->fp, 1 /* push */ );

	while( !retcode )
		retcode = executeInstruction( stp );

	if( stp->fp && stp->fp->ex )
		stp->ex = stp->fp->ex;

	/* restore the current stack frame */
	stp->fp = fp;

	return retcode;
}


int execute( ClassFile* cf, Stack* stp, int argc, char** argv )
{
	ObjRef* args;
	int retcode = 0;
	u2 n;

	for( n = 0; n < cf->methods_count; n++ )
	{
		if( cf->methods[n] && checkSimpleName( cf, cf->methods[n]->name_index, "main" ) )
		{
			break;
		}
	}

	if( n >= cf->methods_count )
	{
		printf( "Can't find method \"main\"\n" );
		newCompleteException( stp, NoSuchMethodError, "\"main\"" );
		return 2 /* exception */;
	}

	if( (cf->methods[n]->access_flags & ACC_NATIVE) || !cf->methods[n]->ca || !cf->methods[n]->ca->code )
	{
		printf( "No Java code for method \"main\"\n" );
		newCompleteException( stp, NoSuchMethodError, "main" );
		return 2 /* exception */;
	}

	stp->fp = newStackFrame( cf->methods[n], NULL );

	if( !cf->initialised )
	{
		INITIALISE( cf, stp->fp );
	}

	args = newObjectRefAndArray( T_OBJREF, argc, sizeof(ObjRef*) );

	if( argc && args )
	{
		int n;

		for( n = 0; n < argc; n++ )
			args->data[n] = (u4) newJavaLangString( (u1*) argv[n], strlen( argv[n] ) );
	}

	PUT_LOBJREF( stp->fp, 0, args );

#ifdef _DEBUG
	if( method_debug )
	{
		char buf[255];
		sprintMethodInfo( buf, cf->methods[n] );
		printf( "Java method %s\n", buf );
	}
#endif

	if( callstack )
		(*callstack)( stp->fp, 2 /* set */ );

	while( !retcode )
		retcode = executeInstruction( stp );

	return retcode;
}
