/******************************************************************************
 *
 * Filename: loader.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Class file loader.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "config.h"
#include "classfile.h"
#include "disasm.h"
#include "util.h"
#include "memory.h"
#include "globals.h"

#include "loader.h"


extern void systemOutPrint( const char* buf, int len );

#ifndef PLATFORM_LINUX_I386
extern char* strdup( const char* s );
#endif


ClassFileRecord* loadedTable = NULL;


/** Function to access loaded table - should not access directly */
ClassFile* checkLoaded( const char *classname )
{
	ClassFileRecord* cfr;
    
	for( cfr = loadedTable; cfr && strcmp( cfr->classname, classname ); cfr = cfr->next );

	if( cfr )
	    return cfr->cf;
	
    return NULL;
}


void readU1( void* d, FILE* f )
{
	u1* data = d;
	*data = fgetc( f );
}

void readU2( void* d, FILE* f )
{
	u1* data = d;
#ifdef CONFIG_LITTLE_ENDIAN
	u1 b1 = fgetc( f );
	u1 b2 = fgetc( f );
	*data = b2; data++;
	*data = b1;
#else
	*data = fgetc( f ); data++;
	*data = fgetc( f );
#endif
}

void readU4( void* d, FILE* f )
{
	u1* data = d;
#ifdef CONFIG_LITTLE_ENDIAN
	u1 b1 = fgetc( f );
	u1 b2 = fgetc( f );
	u1 b3 = fgetc( f );
	u1 b4 = fgetc( f );
	*data = b4; data++;
	*data = b3; data++;
	*data = b2; data++;
	*data = b1;
#else
	*data = fgetc( f ); data++;
	*data = fgetc( f ); data++;
	*data = fgetc( f ); data++;
	*data = fgetc( f );
#endif
}

u4 readN( void* d, FILE* f, u4 count )
{
	return fread( d, 1, count, f );
}


/**
 * if bit7 = 0, then single byte (\u0001 - \u007f)
 * if bits765 = 110, the two bytes, second byte bits76 = 10 (\u0080 - \u07ff)
 * if bits7651 = 1110, then three bytes, second and third bits76 = 10 (\u0800 - \uffff)
 * mask 11000000 = c0  value 10000000 = 80
 * mask 11100000 = e0  value 11000000 = c0
 * mask 11110000 = f0  value 11100000 = e0
 */
u2 countUtf8( u1* buffer, u2 bufflen )
{
	u2 count = 0;
	u2 n;

	for( n = 0; n < bufflen; n++ )
	{
		if( (buffer[n] & 0x80) == 0x00 ) { count++; continue; }	/* ordinary byte */
		if( (buffer[n] & 0xc0) == 0x80 ) { continue; }			/* continuation byte */
		if( (buffer[n] & 0xe0) == 0xc0 ) { count++; continue; }   /* two-byte start byte */
		if( (buffer[n] & 0xf0) == 0xe0 ) { count++; continue; }   /* three-byte start byte */

		printf( "LOADEE: Corrupt Utf8 constant - not counting\n" );
		return bufflen;
	}

	return count;
}

void decodeUtf8( u1* bytes, u1* buffer, u2 bufflen )
{
	u2 count = 0;
	u2 n, value;
	u1 x, y, z;

	for( n = 0; n < bufflen; n++ )
	{
		if( (buffer[n] & 0x80) == 0x00 ) { bytes[count++] = buffer[n]; continue; }	/* ordinary byte */
		if( (buffer[n] & 0xe0) == 0xc0 )   /* two-byte start byte */
		{
			if( (n+1) < bufflen )
			{
				x = buffer[n];
				y = buffer[++n];
				value = ((x & 0x1f) << 6) + (y & 0x3f);
				if( value > 0x00ff ) printf( "WARNING: Utf8 character %04x will be truncated\n", value );
				bytes[count++] = (u1) (value & 0x00ff);
				continue;
			}
		}
		if( (buffer[n] & 0xf0) == 0xe0 )   /* three-byte start byte */
		{
			if( (n+1) < bufflen )
			{
				x = buffer[n];
				y = buffer[++n];
				z = buffer[++n];
				value = ((x & 0xf) << 12) + ((y & 0x3f) << 6) + (z & 0x3f);
				if( value > 0x00ff ) printf( "WARNING: Utf8 character %04x will be truncated\n", value );
				bytes[count++] = (u1) (value & 0x00ff);
				continue;
			}
		}

		printf( "LOADER: Corrupt Utf8 constant - not decoding\n" );
		memcpy( bytes, buffer, bufflen );
		return;
	}
}


cp_info* readCPInfo( FILE* f, int* flag )
{
	cp_info* info = NULL;
	u1 tag;

	readU1( &tag, f );

	switch( tag )
	{
	case CONSTANT_Utf8:
		{
			CONSTANT_Utf8_info* Utf8Info = (CONSTANT_Utf8_info*) javavm_malloc( sizeof( CONSTANT_Utf8_info ) );
			u1* buffer;
			u2 bufflen;

			readU2( &bufflen, f );
			buffer = (u1*) javavm_malloc( bufflen );
			readN( buffer, f, bufflen );
			Utf8Info->length = countUtf8( buffer, bufflen );
			if( Utf8Info->length == bufflen )
			{
				Utf8Info->bytes = buffer;
			}
			else
			{
				if( verbose ) printf( "Decoding Utf8 string\n" );
				Utf8Info->bytes = (u1*) javavm_malloc( Utf8Info->length );
				decodeUtf8( Utf8Info->bytes, buffer, bufflen );
				javavm_free( buffer, bufflen );
			}
			info = (cp_info*) Utf8Info;
		}
		break;
	case CONSTANT_Integer:
	case CONSTANT_Float:
		{
			CONSTANT_Integer_info* IFInfo = (CONSTANT_Integer_info*) javavm_malloc( sizeof( CONSTANT_Integer_info ) );
			readU4( &IFInfo->bytes, f );
			info = (cp_info*) IFInfo;
		}
		break;
	case CONSTANT_Long:
	case CONSTANT_Double:
		{
			CONSTANT_Long_info* LDInfo = (CONSTANT_Long_info*) javavm_malloc( sizeof( CONSTANT_Long_info ) );
			readU4( &LDInfo->high_bytes, f );
			readU4( &LDInfo->low_bytes, f );
			info = (cp_info*) LDInfo;
			*flag = 1;
		}
		break;
	case CONSTANT_Class:
		{
			CONSTANT_Class_info* CInfo = (CONSTANT_Class_info*) javavm_malloc( sizeof( CONSTANT_Class_info ) );
			readU2( &CInfo->name_index, f );
			CInfo->cf = NULL;
			info = (cp_info*) CInfo;
		}
		break;
	case CONSTANT_String:
		{
			CONSTANT_String_info* strInfo = (CONSTANT_String_info*) javavm_malloc( sizeof( CONSTANT_String_info ) );
			readU2( &strInfo->string_index, f );
			strInfo->objref = NULL;
			info = (cp_info*) strInfo;
		}
		break;
	case CONSTANT_Fieldref:
		{
			CONSTANT_Fieldref_info* finfo = (CONSTANT_Fieldref_info*) javavm_malloc( sizeof( CONSTANT_Fieldref_info ) );
			readU2( &finfo->class_index, f );
			readU2( &finfo->name_and_type_index, f );
			finfo->info = NULL;
			info = (cp_info*) finfo;
		}
		break;
	case CONSTANT_Methodref:
		{
			CONSTANT_Methodref_info* minfo = (CONSTANT_Methodref_info*) javavm_malloc( sizeof( CONSTANT_Methodref_info ) );
			readU2( &minfo->class_index, f );
			readU2( &minfo->name_and_type_index, f );
			minfo->info = NULL;
			info = (cp_info*) minfo;
		}
		break;
	case CONSTANT_InterfaceMethodref:
		{
			CONSTANT_InterfaceMethodref_info* iminfo = (CONSTANT_InterfaceMethodref_info*) javavm_malloc( sizeof( CONSTANT_InterfaceMethodref_info ) );
			readU2( &iminfo->class_index, f );
			readU2( &iminfo->name_and_type_index, f );
			iminfo->info = NULL;
			info = (cp_info*) iminfo;
		}
		break;
	case CONSTANT_NameAndType:
		{
			CONSTANT_NameAndType_info* NATInfo = (CONSTANT_NameAndType_info*) javavm_malloc( sizeof( CONSTANT_NameAndType_info ) );
			readU2( &NATInfo->name_index, f );
			readU2( &NATInfo->descriptor_index, f );
			info = (cp_info*) NATInfo;
		}
		break;
	default:
		printf( "LOADER: Bad constant pool item tag (%d)\n", tag );
		return NULL;
	}

	if( info )
		info->tag = tag;

	return info;
}


Exception_table* readException( FILE* f )
{
	Exception_table* ex = (Exception_table*) javavm_malloc( sizeof( Exception_table ) );
	readU2( &ex->start_pc, f );
	readU2( &ex->end_pc, f );
	readU2( &ex->handler_pc, f );
	readU2( &ex->catch_type, f );
	return ex;
}


/* forwards declaration */
attribute_info* readAttribute( ClassFile* cf, FILE* f );


attribute_info* readCodeAttribute( u2 name_index, u4 length, ClassFile* cf, FILE* f )
{
	Code_attribute* ca = (Code_attribute*) javavm_malloc( sizeof( Code_attribute ) );
	u2 n;

	ca->attribute_name_index = name_index;
	ca->attribute_length = length;

	readU2( &ca->max_stack, f );
	readU2( &ca->max_locals, f );
	readU4( &ca->code_length, f );

	if( ca->code_length )
	{
		ca->code = javavm_malloc( ca->code_length );
		if( readN( ca->code, f, ca->code_length ) != ca->code_length )
			return NULL;
	}
	else
		ca->code = NULL;

	readU2( &ca->exception_table_length, f );

	if( ca->exception_table_length )
	{
		ca->exception_table = (Exception_table**) javavm_malloc( ca->exception_table_length * sizeof( Exception_table* ) );

		for( n = 0; n < ca->exception_table_length; n++ )
		{
			if( !(ca->exception_table[n] = readException( f )) )
				return NULL;
		}
	}
	else
		ca->exception_table = NULL;

	readU2( &ca->attributes_count, f );

	if( ca->attributes_count )
	{
		ca->attributes = (attribute_info**) javavm_malloc( ca->attributes_count * sizeof( attribute_info* ) );

		for( n = 0; n < ca->attributes_count; n++ )
		{
			if( !(ca->attributes[n] = readAttribute( cf, f )) )
				return NULL;
		}
	}
	else
		ca->attributes = NULL;

	return (attribute_info*) ca;
}


attribute_info* readLineNumberTableAttribute( u2 name_index, u4 length, ClassFile* cf, FILE* f )
{
	LineNumberTable_attribute* lnta = (LineNumberTable_attribute*) javavm_malloc( sizeof( LineNumberTable_attribute ) ) ;
	u2 n;

	lnta->attribute_name_index = name_index;
	lnta->attribute_length = length;

	readU2( &lnta->line_number_table_length, f );

	if( lnta->line_number_table_length )
	{
		lnta->line_number_table = (LineNumber_table**) javavm_malloc( lnta->line_number_table_length * sizeof( LineNumber_table* ) );

		for( n = 0; n < lnta->line_number_table_length; n++ )
		{
			lnta->line_number_table[n] = (LineNumber_table*) javavm_malloc( sizeof( LineNumber_table ) );
			readU2( &lnta->line_number_table[n]->start_pc, f );	     
			readU2( &lnta->line_number_table[n]->line_number, f );	     
		}
	}
	else
		lnta->line_number_table = NULL;

	return (attribute_info*) lnta;
}


attribute_info* readLocalVariableTableAttribute( u2 name_index, u4 length, ClassFile* cf, FILE* f )
{
	LocalVariableTable_attribute* lvta = (LocalVariableTable_attribute*) javavm_malloc( sizeof( LocalVariableTable_attribute ) ) ;
	u2 n;

	lvta->attribute_name_index = name_index;
	lvta->attribute_length = length;

	readU2( &lvta->local_variable_table_length, f );

	if( lvta->local_variable_table_length )
	{
		lvta->local_variable_table = (LocalVariable_table**) javavm_malloc( lvta->local_variable_table_length * sizeof( LocalVariable_table* ) );

		for( n = 0; n < lvta->local_variable_table_length; n++ )
		{
			lvta->local_variable_table[n] = (LocalVariable_table*) javavm_malloc( sizeof( LocalVariable_table ) );
			readU2( &lvta->local_variable_table[n]->start_pc, f );
			readU2( &lvta->local_variable_table[n]->length, f );
			readU2( &lvta->local_variable_table[n]->name_index, f );
			readU2( &lvta->local_variable_table[n]->descriptor_index, f );
			readU2( &lvta->local_variable_table[n]->index, f );
		}
	}
	else
		lvta->local_variable_table = NULL;

	return (attribute_info*) lvta;
}


attribute_info* readSourceFileAttribute( u2 name_index, u4 length, ClassFile* cf, FILE* f )
{
	SourceFile_attribute* sfa = (SourceFile_attribute*) javavm_malloc( sizeof( SourceFile_attribute ) ) ;

	sfa->attribute_name_index = name_index;
	sfa->attribute_length = length;

	readU2( &sfa->sourcefile_index, f );

	return (attribute_info*) sfa;
}


attribute_info* readAttribute( ClassFile* cf, FILE* f )
{
	attribute_info* attr;
	u2 name_index;
	u4 length;

	readU2( &name_index, f );
	readU4( &length, f );

	if( checkSimpleName( cf, name_index, "Code" ) )
	{
		attr = readCodeAttribute( name_index, length, cf, f );
	}
	else if( checkSimpleName( cf, name_index, "LineNumberTable" ) )
	{
		attr = readLineNumberTableAttribute( name_index, length, cf, f );
	}
	else if( checkSimpleName( cf, name_index, "LocalVariableTable" ) )
	{
		attr = readLocalVariableTableAttribute( name_index, length, cf, f );
	}
	else if( checkSimpleName( cf, name_index, "SourceFile" ) )
	{
		attr = readSourceFileAttribute( name_index, length, cf, f );
	}
	else
	{
		attr = (attribute_info*) javavm_malloc( sizeof( attribute_info ) + length );
		attr->attribute_name_index = name_index;
		attr->attribute_length = length;

		if( attr->attribute_length )
		{
			u1* data = (u1*) attr; data+=sizeof( attribute_info );
			if( readN( data, f, attr->attribute_length ) != attr->attribute_length )
				return NULL;
		}
	}

	return attr;
}


field_info* readField( ClassFile* cf, FILE* f )
{
	field_info* field = (field_info*) javavm_malloc( sizeof( field_info ) );
	u2 n;

	readU2( &field->access_flags, f );
	readU2( &field->name_index, f );
	readU2( &field->descriptor_index, f );
	readU2( &field->attributes_count, f );

	field->cf = NULL;
	field->offset = 0;

	if( field->attributes_count )
	{
		field->attributes = (attribute_info**) javavm_malloc( field->attributes_count * sizeof( attribute_info* ) );

		for( n = 0; n < field->attributes_count; n++ )
		{
			if( !(field->attributes[n] = readAttribute( cf, f )) )
				return NULL;
		}
	}
	else
		field->attributes = NULL;

	return field;
}

method_info* readMethod( ClassFile* cf, FILE* f )
{
	method_info* method = (method_info*) javavm_malloc( sizeof( method_info ) );
	u2 n;

	readU2( &method->access_flags, f );
	readU2( &method->name_index, f );
	readU2( &method->descriptor_index, f );
	readU2( &method->attributes_count, f );

	method->index = 0;
	method->cf = NULL;
	method->args = 0;
	method->breakpoint = 0;
	method->ca = NULL;
	method->lnta = NULL;
	method->lvta = NULL;
	method->nativestr = NULL;

	if( method->attributes_count )
	{
		method->attributes = (attribute_info**) javavm_malloc( method->attributes_count * sizeof( attribute_info* ) );

		for( n = 0; n < method->attributes_count; n++ )
		{
			if( !(method->attributes[n] = readAttribute( cf, f )) )
				return NULL;
		}
	}
	else
		method->attributes = NULL;

	return method;
}


ClassFile* readClassFile( const char* filename, const char* classname, Stack* stp )
{
	CONSTANT_Class_info* ci;
    CONSTANT_Utf8_info* cname;
    ClassFileRecord* cfr;
	ClassFile* cf;
	FILE *f;
	u2 n;
	u4 version;

	if( !(f = fopen( filename, "rb" )) )
	{
		if( verbose )
		{
			char buf[255];
			sprintf( buf, "Failed to open \"%s\"\n", filename );
			systemOutPrint( buf, strlen( buf ) );
		}

		stp->retcode = newCompleteException( stp, NoClassDefFoundError, classname );
		return NULL;
	}

	cf = (ClassFile*) javavm_malloc( sizeof( ClassFile ) );
	cf->super = NULL;
	cf->virtuals_count = 0;
	cf->virtuals = NULL;
	cf->name = NULL;
	cf->sfa = NULL;
	cf->resolved = 0;
	cf->initialised = 0;
	cf->object_size = 0;
	cf->static_size = 0;
	cf->static_data = NULL;

	readU4( &cf->magic, f );

	if( cf->magic != 0xCAFEBABE )
	{
		char buf[255];
		sprintf( buf, "LOADER: \"%s\" is not a Java Class file\n", filename );
		systemOutPrint( buf, strlen( buf ) );
		fclose( f );
		stp->retcode = newCompleteException( stp, ClassFormatError, classname );
		return NULL;
	}

	if( verbose )
	{
		char buf[255];
		sprintf( buf, "Loading class \"%s\"...\n", filename );
		systemOutPrint( buf, strlen( buf ) );
	}

	readU2( &cf->minor_version, f );
	readU2( &cf->major_version, f );

	version = (cf->major_version << 16) | cf->minor_version;

	if( (version < 0x002d0000L) || (version > 0x003f0000L ) )
	{
		fclose( f );
		printf( "LOADER: Aborted load of \"%s\" at version number\n", filename );
		stp->retcode = newCompleteException( stp, ClassFormatError, classname ); /* ### will be UnsupportedClassVersionError in JDK 1.2 */
		return NULL;
	}

	if( verbose )
	{
		if( version < 0x002d0003 ) printf( "WARNING: JDK 1.0.2 class being loaded\n" );
		else if( version >= 0x002e0000 ) printf( "WARNING: JDK 1.2 or later class being loaded\n" );
	}

	readU2( &cf->constant_pool_count, f );

	if( cf->constant_pool_count )
	{
		int flag = 0;

		cf->constant_pool = (cp_info**) javavm_malloc( cf->constant_pool_count * sizeof( cp_info* ) );
		cf->constant_pool[0] = NULL; /* dummy first element */

		for( n = 1; n < cf->constant_pool_count; n++ )
		{
			if( !(cf->constant_pool[n] = readCPInfo( f, &flag )) )
			{
				fclose( f );
				printf( "LOADER: Aborted load of \"%s\" at constant pool\n", filename );
				return NULL;
			}

			/* Integer or Double requires trailing blank constant pool entry */
			if( flag )
			{
				flag = 0;
				cf->constant_pool[n+1] = cf->constant_pool[n];
				n++;
			}
		}
	}
	else
		cf->constant_pool = NULL;

	readU2( &cf->access_flags, f );
	readU2( &cf->this_class, f );

	if( !CHECKCONSTANT( cf, cf->this_class, CONSTANT_Class ) )
	{
		printf( "LOADER: Aborted load of \"%s\" at this class\n", filename );
		stp->retcode = newCompleteException( stp, ClassFormatError, classname );
		return NULL;
	}

	ci = (CONSTANT_Class_info*) cf->constant_pool[cf->this_class];
	ci->cf = cf;

	if( !CHECKCONSTANT( cf, ci->name_index, CONSTANT_Utf8 ) )
	{
		printf( "LOADER: Aborted load of \"%s\" at name\n", filename );
		stp->retcode = newCompleteException( stp, ClassFormatError, classname );
		return NULL;
	}

	cname = (CONSTANT_Utf8_info*) cf->constant_pool[ci->name_index];
	cf->name = (char*) javavm_malloc( cname->length + 1 );
	strncpy( cf->name, (const char*) cname->bytes, cname->length );
	cf->name[ cname->length ] = '\0';

	readU2( &cf->super_class, f );
	readU2( &cf->interfaces_count, f );

	if( cf->interfaces_count )
	{
		cf->interfaces = (InterfaceEntry**) javavm_malloc( cf->interfaces_count * sizeof( InterfaceEntry* ) );

		for( n = 0; n < cf->interfaces_count; n++ )
		{
			cf->interfaces[n] = (InterfaceEntry*) javavm_malloc( sizeof( InterfaceEntry ) );
			readU2( &(cf->interfaces[n]->class_index), f );
			cf->interfaces[n]->cf = NULL;
		}
	}
	else
		cf->interfaces = NULL;

	readU2( &cf->fields_count, f );

	if( cf->fields_count )
	{
		cf->fields = (field_info**) javavm_malloc( cf->fields_count * sizeof( field_info* ) );

		for( n = 0; n < cf->fields_count; n++ )
		{
			if( !(cf->fields[n] = readField( cf, f )) )
			{
				fclose( f );
				printf( "LOADER: Aborted load of \"%s\" at field table\n", filename );
				stp->retcode = newCompleteException( stp, ClassFormatError, classname );
				return NULL;
			}
		}
	}
	else
		cf->fields = NULL;

	readU2( &cf->methods_count, f );

	if( cf->methods_count )
	{
		cf->methods = (method_info**) javavm_malloc( cf->methods_count * sizeof( method_info* ) );

		for( n = 0; n < cf->methods_count; n++ )
		{
			if( !(cf->methods[n] = readMethod( cf, f )) )
			{
				printf( "LOADER: Aborted load of \"%s\" at method table\n", filename );
				fclose( f );
				stp->retcode = newCompleteException( stp, ClassFormatError, classname );
				return NULL;
			}

			cf->methods[n]->index = n;
		}
	}
	else
		cf->methods = NULL;

	readU2( &cf->attributes_count, f );

	if( cf->attributes_count )
	{
		cf->attributes = (attribute_info**) javavm_malloc( cf->attributes_count * sizeof( attribute_info* ) );

		for( n = 0; n < cf->attributes_count; n++ )
		{
			if( !(cf->attributes[n] = readAttribute( cf, f )) )
			{
				printf( "LOADER: Aborted load of \"%s\" at attribute table\n", filename );
				fclose( f );
				stp->retcode = newCompleteException( stp, ClassFormatError, classname );
				return NULL;
			}
		}
	}
	else
		cf->attributes = NULL;

	fclose( f );

	cfr = (ClassFileRecord*) javavm_malloc( sizeof( ClassFileRecord ) );
    cfr->classname = cf->name; /* Share with class name field, as now guarenteed set */
	cfr->cf = cf;
	cfr->next = loadedTable; /* head insert */
	loadedTable = cfr;

	/* Create a Class object for this class - done after tbe add because this looks up java/lang/Class */
	cf->classobj = newJavaLangClass( cf );

	return cf;
}
