/******************************************************************************
 *
 * Filename: java_io_FileInputStream.h
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.io.FileInputStream.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#ifndef WIN32
#include <errno.h>
#endif

#include "../engine/config.h"

#include "java_io_FileInputStream.h"

#include "../engine/util.h"
#include "../engine/globals.h"

 
/******************************************************************************
 *
 * class java.io.FileInputStream
 *
 *
 * The system dependent file descriptor. The value is
 * 1 more than actual file descriptor. This means that
 * the default value 0 indicates that the file is not open.
 *
 * class FileInputStream
 * {
 *		private FileDescriptor fd;
 * };
 *
 *****************************************************************************/


/**
 * Returns the number of bytes that can be read from this file input
 * stream without blocking.
 *
 * @return     the number of bytes that can be read from this file input
 *             stream without blocking.
 * @exception  IOException  if an I/O error occurs.
 * @since      JDK1.0
 *
 * int available()
 *     throws IOException; */
jint Java_java_io_FileInputStream_available( JNIEnv* env, jobject self )
{
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	jint pos, end, avail, fd;

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return 0;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return 0;
	}

	switch( fd - 1 )
	{
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
		return 0;
	case 0: /* stdin */
	default:
		pos = ftell( fdList[fd-1] );
		fseek( fdList[fd-1], 0, SEEK_END );
		end = ftell( fdList[fd-1] );
		avail = end - pos;
		fseek( fdList[fd-1], pos, SEEK_SET );
		return avail;
	}
}


/**
 * Opens the specified file for reading.
 * @param name the name of the file
 *
 * void open(java.lang.String name)
 *     throws IOException; */
void Java_java_io_FileInputStream_open( JNIEnv* env, jobject self, jobject str )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, str ) )
	{
		int fd = findFreeFD();

		if( !(fdList[fd] = fopen( filename, "rb" )) )
		{
			char buf[1024];
#ifdef WIN32
			sprintf( buf, "Failed to open file for reading \"%s\"", filename );
#else
			sprintf( buf, "Failed to open file for reading \"%s\" (errno = %d)", filename, errno );
#endif
			env->retcode = newCompleteException( env, IOException, buf );
			return;
		}

		self->data[0] = (u4) newJavaIoFileDescriptor( fd + 1 );
		return;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
	return;
}


/**
 * Reads a byte of data from this input stream. This method blocks 
 * if no input is yet available. 
 *
 * @return     the next byte of data, or <code>-1</code> if the end of the
 *             file is reached.
 * @exception  IOException  if an I/O error occurs.
 * @since      JDK1.0
 *
 * int read()
 *     throws IOException; */
jint Java_java_io_FileInputStream_read( JNIEnv* env, jobject self )
{
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	u4 fd;
	jint c;

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return 0;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return 0;
	}

	switch( fd - 1 )
	{
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
		return 0;
	case 0: /* stdin */
	default:
		c = fgetc( fdList[fd-1] ); /* != EOF */
		break;
	}

	return c;
}


/** 
 * Reads a subarray as a sequence of bytes. 
 * @param b the data to be written
 * @param off the start offset in the data
 * @param len the number of bytes that are written
 * @exception IOException If an I/O error has occurred. 
 *
 * int readBytes(byte b[], int off, int len)
 *     throws IOException; */
jint Java_java_io_FileInputStream_readBytes( JNIEnv* env, jobject self, jobject array, jint off, jint len )
{
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	u4 fd;

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad file descriptor" );
		return 0;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return 0;
	}

	if( array && (array->cf == NULL) && (array->type == T_BYTE) )
	{
		char* dst = (char*) array->data;
		int c;
		int32 n;

		switch( fd - 1 )
		{
		case 1: /* stdout */
		case 2: /* stderr */
			env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
			return 0;
		case 0: /* stdin */
		default:
			for( n = off; (n < off + len) && ((c = fgetc( fdList[fd-1] )) != EOF); n++ )
			{
				dst[n] = c;
			}
			break;
		}

		return (n - off);
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to read bytes from FileInputStream" );
	return 0;
}


/**
 * Skips over and discards <code>n</code> bytes of data from the 
 * input stream. The <code>skip</code> method may, for a variety of 
 * reasons, end up skipping over some smaller number of bytes, 
 * possibly <code>0</code>. The actual number of bytes skipped is returned.
 *
 * @param      n   the number of bytes to be skipped.
 * @return     the actual number of bytes skipped.
 * @exception  IOException  if an I/O error occurs.
 * @since      JDK1.0
 *
 * long skip( long )
 *     throws IOException; */
jlong Java_java_io_FileInputStream_skip( JNIEnv* env, jobject self, jlong len )
{
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	u4 fd;

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return 0;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode =  newCompleteException( env, IOException, "Bad file descriptor" );	
		return 0;
	}

	switch( fd - 1 )
	{
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
		return 0;
	case 0: /* stdin */
	default:
		fseek( fdList[fd-1], (long) len, SEEK_CUR );
		return len; /* ### assume success! */
	}

}


/**
 * Closes this file input stream and releases any system resources 
 * associated with the stream. 
 *
 * @exception  IOException  if an I/O error occurs.
 * @since      JDK1.0
 *
 * void close()
 *      throws IOException; */
void Java_java_io_FileInputStream_close( JNIEnv* env, jobject self )
{
	ObjRef *fileDesc = (ObjRef*) self->data[0];
	u4 fd;

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad file descriptor" );
		return;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	switch( fd - 1 )
	{
	case 0: /* stdin */
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Close not permitted on in, out or err" );
		return;
	default:
		fclose( fdList[fd-1] );
		fdList[fd-1] = 0;
		return;;
	}
}
