/******************************************************************************
 *
 * Filename: java_io_ObjectStreamClass.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.io.ObjectStreamClass.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "../engine/config.h"

#include "java_io_ObjectStreamClass.h"

#include "../engine/util.h"
#include "../engine/disasm.h"
#include "../engine/globals.h"


/* static int getClassAccess( Class aclass ) */
jint Java_java_io_ObjectStreamClass_getClassAccess( JNIEnv *env, jobject self, jclass aclass )
{
	ClassFile* cf = (ClassFile*) aclass->data;
	return cf->access_flags;
}


/* private static native int getFieldAccess(Class aclass, String fieldsig); */
jint Java_java_io_ObjectStreamClass_getFieldAccess( JNIEnv *env, jclass staticself, jclass aclass, jobject fieldsig )
{
	ClassFile* cf = (ClassFile*) aclass->data;
	char buf[1024];
	const char* name = buf;
	const char* desc = NULL;
	u2 n;

	extractFromJavaLangString( buf, 1024, fieldsig );
	for( n = 0; buf[n]; n++ )
	{
		if( buf[n] == ' ' )
		{
			buf[n] = '\0';
			desc = &buf[n+1];
			break;
		}
	}

	for( n = 0; n < cf->fields_count; n++ )
	{
		if( cf->fields[n]
			&& checkSimpleName( cf, cf->fields[n]->name_index, name )
			&& checkSimpleName( cf, cf->fields[n]->descriptor_index, desc ) )
		{
			break;
		}
	}

	if( n >= cf->fields_count )
	{
		printf( "WARNING: ObjectStreamClass_getFieldAccess() failed - returning 0.\n" );
		return 0;
	}

	return cf->fields[n]->access_flags;
}


/* java.io.ObjectStreamField[] getFields0( java.lang.Class cl ) */
jobject Java_java_io_ObjectStreamClass_getFields0( JNIEnv *env, jobject self, jclass cls )
{
	ClassFile* cf = (ClassFile*) cls->data;
	ObjRef* array = newObjectRefAndArray( T_OBJREF, cf->fields_count, sizeof(ObjRef*) );
	int n;

	addToList( array, &arrayPool );

	for( n = 0; n < cf->fields_count; n++ )
	{
		if( CHECKCONSTANT( cf, cf->fields[n]->name_index, CONSTANT_Utf8 )
			&& CHECKCONSTANT( cf, cf->fields[n]->descriptor_index, CONSTANT_Utf8 ) )
		{
			CONSTANT_Utf8_info* ni = (CONSTANT_Utf8_info*) cf->constant_pool[cf->fields[n]->name_index];
			ObjRef* name = newJavaLangString( ni->bytes, ni->length );
			CONSTANT_Utf8_info* di = (CONSTANT_Utf8_info*) cf->constant_pool[cf->fields[n]->descriptor_index];
			ObjRef* typeString = newJavaLangString( di->bytes, di->length );

			array->data[n] = (u4) newJavaIoObjectStreamField( name, di->bytes[0], cf->fields[n]->offset, typeString );
			addToList( name, &objectPool );
			addToList( typeString, &objectPool );
			addToList( (ObjRef*) array->data[n], &objectPool );
		}
	}

	return array;
}


/* static String[] getFieldSignatures( Class aclass ) */
jobject Java_java_io_ObjectStreamClass_getFieldSignatures( JNIEnv *env, jclass staticself, jclass aclass )
{
	ClassFile* cf = (ClassFile*) aclass->data;
	ObjRef* array = newObjectRefAndArray( T_OBJREF, cf->fields_count, sizeof(ObjRef*) );
	int n;

	addToList( array, &arrayPool );

	for( n = 0; n < cf->fields_count; n++ )
	{
		if( CHECKCONSTANT( cf, cf->fields[n]->name_index, CONSTANT_Utf8 )
			&& CHECKCONSTANT( cf, cf->fields[n]->descriptor_index, CONSTANT_Utf8 ) )
		{
			char buffer[256];
			char* buf = buffer;
			CONSTANT_Utf8_info* ni = (CONSTANT_Utf8_info*) cf->constant_pool[cf->fields[n]->name_index];
			CONSTANT_Utf8_info* di = (CONSTANT_Utf8_info*) cf->constant_pool[cf->fields[n]->descriptor_index];

			strncpy( buf, (const char*) ni->bytes, ni->length ); buf += ni->length;
			strncpy( buf, " ", 1 ); buf += 1;
			strncpy( buf, (const char*) di->bytes, di->length ); buf += di->length;
			strncpy( buf, "\0", 1 );

			array->data[n] = (u4) newJavaLangString( (u1*) buffer, ni->length + di->length + 1 );
			addToList( (ObjRef*) array->data[n], &objectPool );
		}
	}

	return array;
}


/* private static native int getMethodAccess(Class aclass, String fieldsig); */
jint Java_java_io_ObjectStreamClass_getMethodAccess( JNIEnv *env, jclass staticself, jclass aclass, jobject methodsig )
{
	ClassFile* cf = (ClassFile*) aclass->data;
	char buf[1024];
	const char* name = buf;
	const char* desc = NULL;
	u2 n;

	extractFromJavaLangString( buf, 1024, methodsig );
	for( n = 0; buf[n]; n++ )
	{
		if( buf[n] == ' ' )
		{
			buf[n] = '\0';
			desc = &buf[n+1];
			break;
		}
	}

	for( n = 0; n < cf->methods_count; n++ )
	{
		if( cf->methods[n]
			&& checkSimpleName( cf, cf->methods[n]->name_index, name )
			&& checkSimpleName( cf, cf->methods[n]->descriptor_index, desc ) )
		{
			break;
		}
	}

	if( n >= cf->methods_count )
	{
		printf( "WARNING: ObjectStreamClass_getMethodAccess() failed - returning 0.\n" );
		return 0;
	}

	return cf->methods[n]->access_flags;
}


/* static String[] getMethodSignatures( Class aclass ) */
jobject Java_java_io_ObjectStreamClass_getMethodSignatures( JNIEnv *env, jclass staticself, jclass aclass )
{
	ClassFile* cf = (ClassFile*) aclass->data;
	ObjRef* array = newObjectRefAndArray( T_OBJREF, cf->methods_count, sizeof(ObjRef*) );
	int n;

	addToList( array, &arrayPool );

	for( n = 0; n < cf->methods_count; n++ )
	{
		if( CHECKCONSTANT( cf, cf->methods[n]->name_index, CONSTANT_Utf8 )
			&& CHECKCONSTANT( cf, cf->methods[n]->descriptor_index, CONSTANT_Utf8 ) )
		{
			char buffer[256];
			char* buf = buffer;
			CONSTANT_Utf8_info* ni = (CONSTANT_Utf8_info*) cf->constant_pool[cf->methods[n]->name_index];
			CONSTANT_Utf8_info* di = (CONSTANT_Utf8_info*) cf->constant_pool[cf->methods[n]->descriptor_index];

			strncpy( buf, (const char*) ni->bytes, ni->length ); buf += ni->length;
			strncpy( buf, " ", 1 ); buf += 1;
			strncpy( buf, (const char*) di->bytes, di->length ); buf += di->length;
			strncpy( buf, "\0", 1 );

			array->data[n] = (u4) newJavaLangString( (u1*) buffer, ni->length + di->length + 1 );
			addToList( (ObjRef*) array->data[n], &objectPool );
		}
	}

	return array;
}


/* static long getSerialVersionUID( java.lang.Class cl ) */
jlong Java_java_io_ObjectStreamClass_getSerialVersionUID( JNIEnv *env, jclass staticself, jclass cl )
{
	ClassFile* cf = (ClassFile*) cl->data;

	/* ### Implement me! */
	printf( "WARNING: java.io.ObjectStreamClass.getSerialVersionUID() dummied.\n" );

	return 0;
}


/**
 * Get the boolean as to whether the class has/had a writeObject method.
 *
 * private static native boolean hasWriteObject(Class cl)
 */
jboolean Java_java_io_ObjectStreamClass_hasWriteObject( JNIEnv *env, jclass staticself, jclass cl )
{
	ClassFile* cf = (ClassFile*) cl->data;

	/* ### Implement me! */
	printf( "WARNING: java.io.ObjectStreamClass.hasWriteObject() dummied.\n" );

	return 0;
}
