/******************************************************************************
 *
 * Filename: java_lang_Class.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.lang.Class.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "../engine/config.h"

#include "java_lang_Class.h"

#include "../engine/util.h"
#include "../engine/disasm.h"
#include "../engine/globals.h"


/**
 * Returns the <code>Class</code> object associated with the class 
 * with the given string name. 
 * Given the fully-qualified name for a class or interface, this
 * method attempts to locate, load and link the class.  If it
 * succeeds, returns the Class object representing the class.  If
 * it fails, the method throws a ClassNotFoundException.
 * <p>
 * For example, the following code fragment returns the runtime 
 * <code>Class</code> descriptor for the class named 
 * <code>java.lang.Thread</code>: 
 * <ul><code>
 *   Class&nbsp;t&nbsp;= Class.forName("java.lang.Thread")
 * </code></ul>
 *
 * @param      className   the fully qualified name of the desired class.
 * @return     the <code>Class</code> descriptor for the class with the
 *             specified name.
 * @exception  ClassNotFoundException  if the class could not be found.
 * @since      JDK1.0
 *
 * public static native Class forName(String className)
 *	throws ClassNotFoundException;
 */
jclass Java_java_lang_Class_forName( JNIEnv* env, jclass cls, jstring str )
{
	char classname[255];

	if( extractFromJavaLangString( classname, 255, str ) )
	{
		ClassFile* cf;
		u4 n;

		for( n = 0; n < strlen( classname ); n++ )
		{
			if( classname[n] == '.' )
				classname[n] = '/';
		}

		if( !( cf = getClass( env, classname ) ) )
		{
			if( verbose )
				printf( "Class.forName() failed to load class \"%s\"\n", classname );

			/* all exceptions translate into ClassNotFoundException's */
			env->retcode = newCompleteException( env, ClassNotFoundException, classname );
			return NULL;
		}

		return cf->classobj;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
	return NULL;
}


/**
 * Determines the class loader for the class. 
 *
 * @return  the class loader that created the class or interface
 *          represented by this object, or <code>null</code> if the
 *          class was not created by a class loader.
 * @see     java.lang.ClassLoader
 * @since   JDK1.0
 *
 * public native java.lang.ClassLoader getClassLoader()
 */
jobject Java_java_lang_Class_getClassLoader( JNIEnv* env, jobject self )
{
	/* ### we are always the system class loader so return null */
	return NULL;
}


/**
 * Determines the interfaces implemented by the class or interface 
 * represented by this object. 
 * <p>
 * If this object represents a class, the return value is an array 
 * containing objects representing all interfaces implemented by the 
 * class. The order of the interface objects in the array corresponds 
 * to the order of the interface names in the </code>implements</code> 
 * clause of the declaration of the class represented by this object. 
 * <p>
 * If this object represents an interface, the array contains 
 * objects representing all interfaces extended by the interface. The 
 * order of the interface objects in the array corresponds to the 
 * order of the interface names in the </code>extends</code> clause of 
 * the declaration of the interface represented by this object. 
 * <p>
 * If the class or interface implements no interfaces, the method 
 * returns an array of length 0. 
 *
 * @return  an array of interfaces implemented by this class.
 * @since   JDK1.0
 *
 * public native Class[] getInterfaces()
 */
jobject Java_java_lang_Class_getInterfaces( JNIEnv *env, jobject self )
{
	ClassFile* cf = (ClassFile*) self->data;
	ObjRef* array = newObjectRefAndArray( T_OBJREF, cf->interfaces_count, sizeof(ObjRef*) );
	u2 n;

	addToList( array, &arrayPool );

	for( n = 0; n < cf->interfaces_count; n++ )
		array->data[n] = (u4) cf->interfaces[n]->cf->classobj;

	return array;
}


/**
 * Returns the fully-qualified name of the type (class, interface,
 * array, or primitive) represented by this Class object, as a String.
 *
 * @return  the fully qualified name of the class or interface
 *          represented by this object.
 * @since   JDK1.0
 *
 * public native String getName();
 */
jstring Java_java_lang_Class_getName( JNIEnv* env, jobject self )
{
	ObjRef* str;
	ClassFile* cf;
	char buf[255];
	int len;

	cf = (ClassFile*) self->data;
	len = sprintClassName( buf, cf, cf->this_class );
	str = newJavaLangString( (u1*) buf, len );
	addToList( str, &objectPool );

	return str;
}


jclass primitiveClass[T_LAST];


/*
 * Return the Virtual Machine's Class object for the named
 * primitive type.
 *
 * static native java.lang.Class getPrimitiveClass(String name);
 */
jclass Java_java_lang_Class_getPrimitiveClass( JNIEnv* env, jclass cls, jstring name )
{
	char classname[255];

	if( extractFromJavaLangString( classname, 255, name ) )
	{
		u1 kind = T_LAST;

		if( strcmp( classname, "boolean" ) == 0 )
			kind = T_BOOLEAN;
		else if( strcmp( classname, "char" ) == 0 )
			kind = T_CHAR;
		else if( strcmp( classname, "double" ) == 0 )
			kind = T_DOUBLE;
		else if( strcmp( classname, "float" ) == 0 )
			kind = T_FLOAT;
		else if( strcmp( classname, "int" ) == 0 )
			kind = T_INT;
		else if( strcmp( classname, "long" ) == 0 )
			kind = T_LONG;
		else
			printf( "WARNING: Class.getPrimitiveClass( \"%s\" )\n", classname );

		if( !primitiveClass[kind] )
			primitiveClass[kind] = newJavaLangClass( NULL );

		return primitiveClass[kind];
	}

	env->retcode = newCompleteException( env, InternalError, "Class.getPrimitiveClass() failed" );
	return NULL;
}


/**
 * If this object represents any class other than the class 
 * </code>Object</code>, then the object that represents the superclass 
 * of that class is returned. 
 * <p>
 * If this object is the one that represents the class 
 * </code>Object</code> or this object represents an interface, 
 * </code>null</code> is returned. 
 *
 * @return  the superclass of the class represented by this object.
 * @since   JDK1.0
 *
 * public native java.lang.Class getSuperclass();
 */
jclass Java_java_lang_Class_getSuperclass( JNIEnv* env, jobject self )
{
	ClassFile* cf = (ClassFile*) self->data;

	if( !cf || !cf->super || (cf->access_flags & ACC_INTERFACE) )
		return NULL;

	return cf->super->classobj;
}


/**
 * If this Class object represents an array type, returns true,
 * otherwise returns false.
 *
 * @since   JDK1.1
 *
 * public native boolean isArray();
 */
jboolean Java_java_lang_Class_isArray( JNIEnv* env, jobject self )
{
	/* ### At the moment all Class objects are not arrays */
	return 0;
}


/**
 * Determines if the class or interface
 * represented by this Class object is either the same as, or is a
 * superclass or superinterface of, the class or interface
 * represented by the specified Class parameter. It returns true
 * if so, false otherwise. If this Class object represents a
 * primitive type, returns true if the specified Class parameter
 * is exactly this Class object, false otherwise.
 *
 * <p>Specifically, this method tests whether the type represented
 * by the specified Class parameter can be converted to the type
 * represented by this Class object via an identity conversion or
 * via a widening reference conversion. See <em>The Java Language
 * Specification</em>, sections 5.1.1 and 5.1.4 , for details.
 *
 * @exception NullPointerException if the specified Class parameter is null.
 * @since   JDK1.1
 *
 * public native boolean isAssignableFrom( java.lang.Class cls )
 */
jboolean Java_java_lang_Class_isAssignableFrom( JNIEnv* env, jobject self, jclass cls )
{
	ClassFile* selfcf = (ClassFile*) self->data;
	ClassFile* clscf  = (ClassFile*) cls->data;

	if( subclassOf( clscf, selfcf ) )
		return 1;

	return 0;
}
