/******************************************************************************
 *
 * Filename: java_lang_Object.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.lang.Object.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../engine/config.h"

#include "java_lang_Object.h"

#include "../engine/disasm.h"
#include "../engine/memory.h"
#include "../engine/util.h"

/**
 * Creates a new object of the same class as this object. It then 
 * initializes each of the new object's fields by assigning it the 
 * same value as the corresponding field in this object. No 
 * constructor is called. 
 * <p>
 * The <code>clone</code> method of class <code>Object</code> will 
 * only clone an object whose class indicates that it is willing for 
 * its instances to be cloned. A class indicates that its instances 
 * can be cloned by declaring that it implements the 
 * <code>Cloneable</code> interface. 
 *
 * @return     a clone of this instance.
 * @exception  CloneNotSupportedException  if the object's class does not
 *               support the <code>Cloneable</code> interface. Subclasses
 *               that override the <code>clone</code> method can also
 *               throw this exception to indicate that an instance cannot
 *               be cloned.
 * @exception  OutOfMemoryError            if there is not enough memory.
 * @see        java.lang.Cloneable
 * @since      JDK1.0
 *
 * protected native Object clone() throws CloneNotSupportedException;
 */
jobject Java_java_lang_Object_clone( JNIEnv* env, jobject self )
{
	ObjRef *clone;

	/* ### should throw CloneNotSupportedException if not a subclass
	 * ### of Clonable */

	clone = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
	clone->cf     = self->cf;
	clone->type   = self->type;
	clone->length = self->length;

	if( self->data )
	{
		char buf[255];
		sprintClassName( buf, self->cf, self->cf->this_class );
		printf( "Cloning instance of %s (%d bytes).\n", buf, (int) (4*clone->cf->object_size) );

		clone->data = (u4*) javavm_malloc( 4*clone->cf->object_size );
		memcpy( clone->data, self->data, 4*clone->cf->object_size );
	}
	else
		clone->data = 0;

	return clone;
}


/**
 * Returns the runtime class of an object. 
 *
 * @return  the object of type <code>Class</code> that represents the
 *          runtime class of the object.
 * @since   JDK1.0
 *
 * public final native Class getClass();
 */
jclass Java_java_lang_Object_getClass( JNIEnv* env, jobject self )
{
	return self->cf->classobj;
}


/**
 * Returns a hash code value for the object. This method is 
 * supported for the benefit of hashtables such as those provided by 
 * <code>java.util.Hashtable</code>. 
 * <p>
 * The general contract of <code>hashCode</code> is: 
 * <ul>
 * <li>Whenever it is invoked on the same object more than once during 
 *     an execution of a Java application, the <code>hashCode</code> method 
 *     must consistently return the same integer. This integer need not 
 *     remain consistent from one execution of an application to another 
 *     execution of the same application. 
 * <li>If two objects are equal according to the <code>equals</code> 
 *     method, then calling the <code>hashCode</code> method on each of the 
 *     two objects must produce the same integer result. 
 * </ul>
 *
 * @return  a hash code value for this object.
 * @see     java.lang.Object#equals(java.lang.Object)
 * @see     java.util.Hashtable
 * @since   JDK1.0
 *
 * public native int hashCode();
 */
jint Java_java_lang_Object_hashCode( JNIEnv* stp, jobject self )
{
	/* Hashcode is address of object's data */
	return (jint) self->data ;
}


/**
 * Wakes up all threads that are waiting on this object's monitor. A 
 * thread waits on an object's monitor by calling one of the 
 * <code>wait</code> methods.
 * <p>
 * This method should only be called by a thread that is the owner 
 * of this object's monitor. See the <code>notify</code> method for a 
 * description of the ways in which a thread can become the owner of 
 * a monitor. 
 *
 * @exception  IllegalMonitorStateException  if the current thread is not
 *               the owner of this object's monitor.
 * @see        java.lang.Object#notify()
 * @see        java.lang.Object#wait()
 * @since      JDK1.0
 *
 * public final native void notifyAll();
 */
void Java_java_lang_Object_notifyAll( JNIEnv* stp, jobject self )
{
	printf( "WARNING: Object.notifyAll() dummied.\n" );
}


/**
 * Waits to be notified by another thread of a change in this object.
 * <p>
 * The current thread must own this object's monitor. The thread 
 * releases ownership of this monitor and waits until either of the 
 * following two conditions has occurred: 
 * <ul>
 * <li>Another thread notifies threads waiting on this object's monitor 
 *     to wake up either through a call to the <code>notify</code> method 
 *     or the <code>notifyAll</code> method. 
 * <li>The timeout period, specified by the <code>timeout</code> 
 *     argument in milliseconds, has elapsed. 
 * </ul>
 * <p>
 * The thread then waits until it can re-obtain ownership of the 
 * monitor and resumes execution. 
 * <p>
 * This method should only be called by a thread that is the owner 
 * of this object's monitor. See the <code>notify</code> method for a 
 * description of the ways in which a thread can become the owner of 
 * a monitor. 
 *
 * @param      timeout   the maximum time to wait in milliseconds.
 * @exception  IllegalArgumentException      if the value of timeout is
 *		     negative.
 * @exception  IllegalMonitorStateException  if the current thread is not
 *               the owner of the object's monitor.
 * @exception  InterruptedException          if another thread has
 *               interrupted this thread.
 * @see        java.lang.Object#notify()
 * @see        java.lang.Object#notifyAll()
 * @since      JDK1.0
 *
 * public final native void wait(long timeout) throws InterruptedException;
 */
void Java_java_lang_Object_wait( JNIEnv* env, jobject self, jlong timeout )
{
	env->retcode = newCompleteException( env, InternalError, "Threads not supported." );
}
