/******************************************************************************
 *
 * Filename: java_lang_SecurityManager.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  25th March 2002
 *
 * Purpose:  Native method implementations for java.lang.SecurityManager.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "../engine/config.h"
#include "../engine/util.h"
#include "../engine/globals.h"

#include "java_lang_SecurityManager.h"


/**
 * Returns the stack depth of the most recently executing method 
 * from a class defined using a class loader. 
 *
 * @return  the depth on the stack frame of the most recent occurrence of a
 *          method from a class defined using a class loader; returns
 *          <code>-1</code> if there is no occurrence of a method from
 *          a class defined using a class loader.
 * @since   JDK1.0
 *
 * protected native int classLoaderDepth();
 */
jint Java_java_lang_SecurityManager_classLoaderDepth( JNIEnv* env, jobject self )
{
	/* ### we don't use class loaders yet so return -1 */
	return -1;
}


/**
 * Returns an object describing the most recent class loader executing
 * on the stack. 
 *
 * @return  the class loader of the most recent occurrence on the stack
 *          of a method from a class defined using a class loader;
 *          returns <code>null</code> if there is no occurrence on the
 *          stack of a method from a class defined using a class loader.
 * @since   JDK1.0
 *
 * protected native ClassLoader currentClassLoader();
 */
jobject Java_java_lang_SecurityManager_currentClassLoader( JNIEnv* env, jobject self )
{
	/* ### we don't use class loaders yet so return null */
	return NULL;
}


/**
 * Returns the current execution stack as an array of classes. 
 * <p>
 * The length of the array is the number of methods on the execution 
 * stack. The element at index <code>0</code> is the class of the 
 * currently executing method, the element at index <code>1</code> is 
 * the class of that method's caller, and so on. 
 *
 * @return  the execution stack.
 * @since   JDK1.0
 *
 * protected native Class[] getClassContext();
 */
jobject Java_java_lang_SecurityManager_getClassContext( JNIEnv* env, jobject self )
{
	Frame* next;
	ObjRef* array;
	int n;

	/* count stack frames */
	for( n = 0, next = env->fp; next; next = next->next )
		n++;

	array = newObjectRefAndArray( T_OBJREF, n, sizeof(ObjRef*) );
	addToList( array, &arrayPool );

	/* Build array of classes */
	for( n = 0, next = env->fp; next; next = next->next )
		array->data[n++] = (u4) next->cf->classobj;

	return array;
}
