/******************************************************************************
 *
 * Filename: java_lang_Throwable.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  25th March 2002
 *
 * Purpose:  Native method implementations for java.lang.Throwable.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../engine/config.h"
#include "../engine/disasm.h"
#include "../engine/globals.h"
#include "../engine/util.h"
#include "../engine/memory.h"

#include "java_lang_Throwable.h"


/**
 * Fills in the execution stack trace. This method is useful when an 
 * application is re-throwing an error or exception. For example: 
 * <p><blockquote><pre>
 *     try {
 *         a = b / c;
 *     } catch(ArithmeticThrowable e) {
 *         a = Number.MAX_VALUE;
 *         throw e.fillInStackTrace();
 *     }
 * </pre></blockquote>
 *
 * @return  this <code>Throwable</code> object.
 * @see     java.lang.Throwable#printStackTrace()
 * @since   JDK1.0
 *
 * public native Throwable fillInStackTrace();
 */
extern jobject Java_java_lang_Throwable_fillInStackTrace( JNIEnv* env, jobject self )
{
	Frame* baseFrame = env->fp;
	Frame* fp;
	char* buffer = (char*) javavm_malloc( 8096 );
	char *buf = buffer;
	int len;

	/* Add each of the frames */
	for( fp = baseFrame; fp; fp = (fp->next ? fp->next : fp->init) )
	{
		len = sprintf( buf, "\tat " ); buf += len;

		if( (fp->next != NULL) && (fp->init != NULL) )
		{
			len = sprintf( buf, "STACK CORRUPT!\n" ); buf += len;
			break;
		}

		if( !fp->mi || !fp->mi->cf )
		{
			len = sprintf( buf, "\n" ); buf += len;
			continue;
		}

		len = sprintClassName( buf, fp->mi->cf, fp->mi->cf->this_class ); buf += len;
		len = sprintf( buf, "." ); buf += len;

		if( CHECKCONSTANT( fp->mi->cf, fp->mi->name_index, CONSTANT_Utf8 ) )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->name_index ); buf += len;
		}

		len = sprintf( buf, "(" ); buf += len;

		if( fp->mi->cf->sfa )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->cf->sfa->sourcefile_index ); buf += len;
		}

		if( fp->mi->lnta )
		{
			int pc = fp->pc - fp->min_pc;
			int best = 0;
			int bestn = fp->mi->lnta->line_number_table_length;
			int n;

			for( n = 0; n < fp->mi->lnta->line_number_table_length; n++ )
			{
				if( (pc >= fp->mi->lnta->line_number_table[n]->start_pc)
					&& (fp->mi->lnta->line_number_table[n]->start_pc > best) )
				{
					best = fp->mi->lnta->line_number_table[n]->start_pc;
					bestn = n;
				}
			}

			if( bestn < fp->mi->lnta->line_number_table_length )
			{
				len = sprintf( buf, ":%d", fp->mi->lnta->line_number_table[bestn]->line_number ); buf += len;
			}
		}
		
		/* extra position info giving exact pc of exception */
		if( verbose )
		{
			int pc = fp->pc - fp->min_pc;
			len = sprintf( buf, "#%d", pc ); buf += len;
		}

		len = sprintf( buf, ")\n" ); buf += len;
	}

	if( self->data )
		self->data[0] = (u4) buffer;

	return self;
}


/**
 * The given object must have a void println(char[]) method
 *
 * private native void printStackTrace0(Object s);
 */
extern void Java_java_lang_Throwable_printStackTrace0( JNIEnv* env, jobject self, jobject stream )
{
	u1 *trace = (u1*) self->data[0];
	env->retcode = 0;

	if( trace )
	{
		ObjRef* str = newJavaLangString( trace, strlen((const char*)trace) );
		PUSH_OBJREF( env->fp, stream );
		PUSH_OBJREF( env->fp, str );
		env->retcode = forceExecuteJavaMethod( stream, env, "write", "(Ljava/lang/String;)V" );
	}
	else
	{
		ObjRef* str = newJavaLangString( (u1*) "null\n", 5 );
		PUSH_OBJREF( env->fp, stream );
		PUSH_OBJREF( env->fp, str );
		env->retcode = forceExecuteJavaMethod( stream, env, "write", "(Ljava/lang/String;)V" );
	}
}
